# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains the ansXpl class."""
import json
import pathlib
import random
import string
import weakref

from ansys.api.mapdl.v0 import mapdl_pb2
import numpy as np

from .common_grpc import ANSYS_VALUE_TYPE
from .errors import MapdlRuntimeError


def id_generator(size=6, chars=string.ascii_uppercase):
    """Generate a random string using only uppercase letters."""
    return "".join(random.choice(chars) for _ in range(size))


MYCTYPE = {
    np.int32: "I",
    np.int64: "L",
    np.single: "F",
    np.double: "D",
    np.complex64: "C",
    np.complex128: "Z",
}


class ansXpl:
    """
    ANSYS database explorer.

    Examples
    --------
    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> xpl = mapdl.xpl

    Open a mode file and extract a vector.

    >>> xpl.open('file.mode')
    >>> vec = xpl.read('MASS')
    >>> vec.asarray()
    array([ 4,  7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 40, 43,
            46, 49, 52, 55, 58,  1], dtype=int32)

    """

    def __init__(self, mapdl):
        """Initialize the class."""
        from ansys.mapdl.core.mapdl_grpc import MapdlGrpc

        if not isinstance(mapdl, MapdlGrpc):  # pragma: no cover
            raise TypeError("Must be initialized using MapdlGrpc class")

        self._mapdl_weakref = weakref.ref(mapdl)
        self._filename = None
        self._open = False

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl."""
        return self._mapdl_weakref()

    def open(self, filename, option=""):
        """
        Open an MAPDL file to explore.

        Parameters
        ----------
        filename : str
            Name of the file to open.

        Returns
        -------
        str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open('file.mode')
        ===============================================
        =====      ANSYS File Xplorer            ======
        ===============================================

        Opening the file.mode ANSYS File

        """
        self._filename = filename
        out = self._mapdl.run(f"*XPL,OPEN,{filename},,{option}")
        self._open = True
        return out

    def close(self):
        """
        Close the MAPDL file after opening.

        Returns
        -------
        str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open("file.mode")
        >>> xpl.close()
        =====      ANSYS File Xplorer : Close the file.mode ANSYS File
        """
        response = self._mapdl.run("*XPL,CLOSE")
        self._check_ignored(response)
        self._open = False
        return response

    def list(self, nlev=1):
        """
        List the records at the current level.

        Parameters
        ----------
        nlev: int
            Number of levels to recursively explore.

        Returns
        -------
        str
            Listing of records from the current level.

        Examples
        --------
        Open a full file and list the current records.

        >>> xpl.open("file.full")
        >>> xpl.list()
        =====      ANSYS File Xplorer : List Blocks in File file.full
         ::FULL::HEADER         Size =        652  B     Total  Size =    180.297 KB
         ::FULL::DOFSBYNOD            Size =         24  B
         ::FULL::BACK                 Size =        336  B

         ::FULL::STIFF::HEADER        Size =    117.316 KB
         ::FULL::RHS                  Size =      1.910 KB
         ::FULL::DIAGK                Size =      1.910 KB
         ::FULL::SCLK                 Size =      1.910 KB
         ::FULL::MRK                  Size =        984  B
         ::FULL::NODEEXT              Size =        336  B
         ::FULL::PCGDOFS              Size =        984  B
         ::FULL::BCDOFS               Size =        984  B
         ::FULL::BCVALUES             Size =         12  B

         ::FULL::MASS::HEADER         Size =     50.801 KB
         ::FULL::DIAGM                Size =      1.910 KB
         ::FULL::NGPH                 Size =        336  B

        """
        response = self._mapdl.run(f"*XPL,LIST,{nlev}")
        self._check_ignored(response)
        return response

    def _check_ignored(self, response):
        """Check for ignored in response."""
        if "ignored" in response:
            raise MapdlRuntimeError(response)

    def help(self):
        """
        XPL help message.

        Examples
        --------
        >>> print(xpl.help())
        """
        return self._mapdl.run("*XPL,HELP")

    def step(self, where):
        """
        Go down in the tree of records

        Parameters
        ----------
        where : str
            Path to follow. This path can be composed of several
            levels, for example ``"BRANCH1::SUBBRANCH2::.."``

        Returns
        -------
        str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.step('MASS')
        >>> print(xpl.where())
         =====      ANSYS File Xplorer : Display Current Location
         Current Location : FULL::MASS
            File Location : 7644
        """
        response = self._mapdl.run(f"*XPL,STEP,{where}")
        if "Not Found" in response:
            raise MapdlRuntimeError(response.strip())
        return response

    def info(self, recname, option=""):
        """
        Gives details on a specific record, or all records (using ``"*"``)

        Parameters
        ----------
        recname : str
            Record of interest

        option : str
            Options string.

        Returns
        -------
        str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open('file.full')
        >>> print(xpl.info('NGPH'))
        =====      ANSYS File Xplorer : Information about Block NGPH
        ::NGPH                 Size =      6.289 KB
                 - Record Size   : 81
                 - Data type     : integer values
        """
        return self._mapdl.run(f"*XPL,INFO,{recname},{option}")

    def print(self, recname):
        """
        Print values of a given records, or all records (using ``"*"``).

        Parameters
        ----------
        recname : str
            Record of interest

        option : str
            Options string.

        Returns
        -------
        str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open('file.full')
        >>> print(xpl.print('DOFSBYNOD'))
        =====      ANSYS File Xplorer : Print Block DOFSBYNOD
        DOFSBYNOD :
        Size : 3
               1         2         3

        """
        return self._mapdl.run(f"*XPL,PRINT,{recname}")

    def json(self):
        """
        Return a JSON representation of the tree or records.

        Examples
        --------
        >>> xpl.json()
        {'name': 'FULL',
         'children': [{'name': 'DOFSBYNOD', 'size': 24},
         {'name': 'BACK', 'size': 336},
         {'name': 'STIFF', 'size': 120132},
         {'name': 'RHS', 'size': 1956},
         {'name': 'DIAGK', 'size': 1956},
         {'name': 'SCLK', 'size': 36},
         {'name': 'NODEEXT', 'size': 32},
         {'name': 'PCGDOFS', 'size': 984},
         {'name': 'BCDOFS', 'size': 984},
         {'name': 'BCVALUES', 'size': 20},
         {'name': 'MASS', 'size': 52020},
         {'name': 'DIAGM', 'size': 1236},
         {'name': 'NGPH', 'size': 6440}]}
        """
        self._mapdl.run("*XPL,JSON,_mylocal_.json")
        text = self._mapdl._download_as_raw("_mylocal_.json").decode()
        return json.loads(text)

    def where(self):
        """
        Returns the current location in the MAPDL file.

        Returns
        -------
        str
            String containing the current location.

        Examples
        --------
        >>> print(xpl.where())
         =====      ANSYS File Xplorer : Display Current Location
         Current Location : FULL
            File Location : 412
        """
        return self._mapdl.run("*XPL,WHERE")

    def up(self, nlev=1):
        """
        Go up in the tree.

        nlev : int
            Number of levels to recursively go up, or TOP

        Examples
        --------
        >>> print(xpl.up())
         =====      ANSYS File Xplorer : Go up to 1 level(s)
                     -> Already at the top level. Command is ignored
        """
        if str(nlev).upper().strip() == "TOP":
            return self._mapdl.run("*XPL,UP,TOP")
        return self._mapdl.run(f"*XPL,UP,{nlev}")

    def goto(self, path):
        """
        Go directly to a new location in the file.

        Parameters
        ----------
        path : str
            Absolute path to the new location.

        Examples
        --------
        >>> print(xpl.goto('MASS'))
         =====      ANSYS File Xplorer : Go up to top level(s)
         =====      ANSYS File Xplorer : Step into Block MASS
        """
        return self._mapdl.run(f"*XPL,GOTO,{path}")

    def copy(self, newfile, option=""):
        """
        Copy the current opened as a new file.

        Parameters
        ----------
        newfile : str
            Name of the new file to create

        option: str
            Option.

        Examples
        --------
        >>> xpl.copy('tmpfile.full')
         =====      ANSYS File Xplorer : Copy file.full ANSYS file to file tmpfile.full
            >>      Remove existing output file tmpfile.full
        """
        return self._mapdl.run(f"*XPL,COPY,{newfile},{option}")

    def save(self):
        """Save the current file, ignoring the marked records."""
        response = self._mapdl.run("*XPL,SAVE").strip()
        self._check_ignored(response)
        return response

    def extract(self, recordname, sets="ALL", asarray=False):
        """Import a Matrix/Vector from a MAPDL result file.

        At the moment, this only supports reading the displacement vectors from
        a result file.

        Parameters
        ----------
        recordname : str
            Record name. Currently only supports the ``"NSL"`` record,
            displacement vectors.

        sets : str or int
            Number of sets. Can be ``"ALL"`` or the number of sets to load.

        asarray : bool, optional
            Return a :class:`numpy.ndarray` rather than a :class:`AnsMath
            <ansy.math.core.math.AnsMath>`. Default ``False``.

        Returns
        -------
        numpy.ndarray or ansys.math.core.math.AnsMath
            A :class:`numpy.ndarray` or :class:`AnsMath
            <ansys.math.core.math.AnsMath>` of the displacement vectors,
            depending on the value of ``asarray``.

        Notes
        -----
        This only works on the ``"NSL"`` record of MAPDL result files.

        Examples
        --------
        First, open a result file and extract the displacement vectors for all
        sets.

        >>> xpl.open("file.rst")
        >>> mat = xpl.extract("NSL")
        >>> mat
        Dense APDLMath Matrix (243, 10)

        Convert to a dense numpy array

        >>> arr = mat.asarray()
        >>> arr
        array([[-9.30806802e-03, -2.39600770e-02, -5.37856729e-03, ...,
                -5.61188243e-03, -7.17686067e-11,  3.71893252e-03],
               [-1.60960014e-02,  2.00410618e-02,  8.05822565e-03, ...,
                -1.26917511e-02, -5.14133724e-11, -1.38783485e-03],
               [ 2.54040694e-02,  3.91901513e-03, -2.67965796e-03, ...,
                -1.46365178e-02,  8.31735188e-11, -2.33109771e-03],
               ...,
               [-2.80679551e-03, -1.45686692e-02,  8.05466291e-03, ...,
                 5.88196684e-03,  1.72211103e-02,  6.10079082e-03],
               [-7.06675717e-03,  1.30455037e-02, -6.31685295e-03, ...,
                 1.08619340e-02, -1.72211102e-02,  2.52199472e-03],
               [ 2.29726170e-02,  3.54392176e-03, -1.87020162e-03, ...,
                 1.20642736e-02,  2.58299321e-11,  9.14504940e-04]])

        """
        from ansys.math.core.math import AnsMath

        if recordname.upper() != "NSL":
            raise ValueError("Currently, the only supported recordname is 'NSL'")

        rand_name = id_generator()
        self._mapdl._log.info(
            "Calling MAPDL to extract the %s matrix from %s",
            recordname,
            self._filename,
        )
        num_first = 1
        num_last = 1
        if sets == "ALL":
            num_last = -1

        dtype = np.double
        file_extension = pathlib.Path(self._filename).suffix[1:]
        if file_extension.lower() != "rst":
            raise MapdlRuntimeError(
                "This method only supports extracting records from result files"
            )

        self._mapdl.run(
            f"*DMAT,{rand_name},{MYCTYPE[dtype]},IMPORT,{file_extension},{self._filename},"
            f"{num_first},{num_last},{recordname}",
            mute=False,
        )

        return AnsMath(self._mapdl).mat(dtype=dtype, name=rand_name)

    def read(self, recordname, asarray=False):
        """
        Read a record and return either an APDL math matrix or an APDL math vector.

        Returns
        -------
        ansys.mapdl.AnsMath or ansys.mapdl.AnsVec
            A handle to the APDLMath object.

        asarray : bool, optional
            Return a :class:`numpy.ndarray` rather than a :class:`AnsMath
            <ansys.math.core.math.AnsMath>`. Default ``False``.

        Examples
        --------
        >>> vec = xpl.read('MASS')
        >>> vec.asarray()
        array([ 4,  7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 40, 43,
               46, 49, 52, 55, 58,  1], dtype=int32)
        >>> vec = xpl.read('MASS', asarray=True)
        array([ 4,  7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 40, 43,
               46, 49, 52, 55, 58,  1], dtype=int32)
        """
        from ansys.math.core.math import AnsMath

        rand_name = id_generator()
        response = self._mapdl.run(f"*XPL,READ,{recordname},{rand_name}")
        self._check_ignored(response)
        data_info = self._mapdl._data_info(rand_name)

        dtype = ANSYS_VALUE_TYPE[data_info.stype]
        if dtype is None:
            raise ValueError("Unknown MAPDL data type")

        # return either vector or matrix type
        mm = AnsMath(self._mapdl)
        if data_info.objtype == mapdl_pb2.DataType.VEC:
            out = mm.vec(dtype=dtype, name=rand_name)
        elif data_info.objtype in [
            mapdl_pb2.DataType.DMAT,
            mapdl_pb2.DataType.SMAT,
        ]:  # pragma: no cover
            out = mm.mat(dtype=dtype, name=rand_name)
        else:  # pragma: no cover
            raise ValueError(f"Unhandled MAPDL matrix object type {data_info.objtype}")

        if asarray:
            out = out.asarray()
        return out

    def write(self, recordname, vecname):
        """
        Write a given record back to an MAPDL file.

        Use the write function at your own risk, you may corrupt an existing
        file by changing the size of a record in the file.  This method must be
        used only on a non-compressed file.

        Parameters
        ----------
        recordname : str
            Name of the record you want to overwrite. Your position
            in the file must be set accordingly to this record location
            (same as if you want to read it).

        vecname : str
            Name of the APDLMath vector you want to write in the MAPDL
            file. Its size must be consistent with the existing record.

        Returns
        -------
        str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.write('MASS', vecname)
        """
        response = self._mapdl.run(f"*XPL,WRITE,{recordname},{vecname}")
        self._check_ignored(response)
        return response

    def __repr__(self):
        txt = "MAPDL File Explorer\n"
        if self._open:
            txt += "\tOpen file:%s" % self._filename
            txt += "\n".join(self.where().splitlines()[1:])
        else:
            txt += "\tNo open file"
        return txt



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class TwoDTo3DAnalysis:
    def map2dto3d(self, action="", ldstep="", sbstep="", option="", **kwargs):
        """Initiates a 2-D to 3-D analysis and maps variables.

        APDL Command: MAP2DTO3D

        Parameters
        ----------
        action
            The 2-D to 3-D action to perform:

            START - Start the analysis process by rebuilding the 2-D analysis database (.db) based
                    on the specified load step and substep information, and
                    update nodes to their deformed positions in the 2-D mesh.

            FINISH - Maps solution variables from the 2-D mesh to the extruded 3-D mesh.

        ldstep
            The load step number at which 2-D to 3-D analysis should occur. The
            default value is the highest load step number found in the
            Jobname.Rnnn files (for the current jobname and in the current
            directory).

        sbstep
            The substep number of the specified load step (LDSTEP) at which the
            2-D to 3-D analysis should occur. The default value is the highest
            substep number found in the specified load step in the Jobname.Rnnn
            files (for the current jobname and in the current directory).

        option
            Mapping option:

            (Blank) - Transfer and map all applied boundary conditions, nodal temperatures, loads,
                      and surface pressures from the 2-D mesh to the extruded
                      3-D mesh. This behavior is the default.

            NOBC - No applied boundary conditions or loads are transferred from the 2-D mesh to
                   the extruded 3-D mesh. Nodal temperatures (defined via the
                   BF,TEMP command) are transferred.

        Notes
        -----
        The MAP2DTO3D command initiates the 2-D to 3-D analysis process, sets
        analysis options, rebuilds the database, and maps the solution
        variables from the 2-D mesh to the 3-D mesh.

        Before issuing this command, clear the database (/CLEAR).

        The LDSTEP and SBSTEP values apply only when Action = START.

        For more information, see 2-D to 3-D Analysis in the Advanced Analysis
        Guide.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MAP2DTO3D,{action},{ldstep},{sbstep},{option}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class MacroFiles:
    def cfclos(self, **kwargs):
        """Closes the "command" file.

        APDL Command: ``*CFCLOS``

        Notes
        -----
        This command is valid in any processor.
        """
        command = f"*CFCLOS,"
        return self.run(command, **kwargs)

    def cfopen(self, fname="", ext="", loc="", **kwargs):
        """Opens a "command" file.

        APDL Command: ``*CFOPEN``

        .. warning::
           This command must be run using :func:`non_interactive
           <ansys.mapdl.core.Mapdl.non_interactive>`

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        loc
            Determines whether existing file will be overwritten or appended:

            (blank) : The existing file will be overwritten.
            APPEND :  The file will be appended to the existing file.

        Notes
        -----
        Data processed with the ``*VWRITE`` command will also be written to this
        file if the file is open when the ``*VWRITE`` command is issued.

        This command is valid in any processor.
        """
        return self.run(f"*CFOPEN,{fname},{ext},,{loc}", **kwargs)

    def cfwrite(self, command="", **kwargs):
        """Writes an ANSYS command (or similar string) to a "command" file.

        APDL Command: ``*CFWRITE``

        Parameters
        ----------
        command
            Command or string to be written.  The standard command form of a
            label followed by arguments separated by commas is assumed.
            Command may be a parameter assignment (e.g.,  ``*CFWRITE, A = 5``).

        Notes
        -----
        Writes an ANSYS command (or similar string) to the file opened
        with ``*CFOPEN``.  The Command string is not executed (except
        that numeric and character parameter substitution and
        operations (with embedded ``*``, /, >,$ etc. characters) are
        performed before writing).  When used with ``*GET`` results
        and parameter substitution, an ANSYS command can be created
        from results and then read back into the ANSYS program (or
        used elsewhere).  For example, if the command
        ``*CFWRITE,BF,NNUM,TEMP,TVAL`` is used in a do-loop, where
        TVAL is a parameter value returned from the ``*GET`` operation
        and NNUM is a specified or returned parameter value, a series
        of BF commands, with numerical values substituted for the two
        parameters, will be written.  To create a file without
        parameter substitution, use ``*CREATE``.

        This command is valid in any processor.
        """
        command = f"*CFWRITE,{command}"
        return self.run(command, **kwargs)

    def create(self, fname="", ext="", **kwargs):
        """Opens (creates) a macro file.

        APDL Command: ``*CREATE``

        .. warning::
           This command must be run using :func:`non_interactive
           <ansys.mapdl.core.Mapdl.non_interactive>`

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        See the ``*USE`` command for a discussion of macros.  All
        commands following the ``*CREATE`` command, up to the ``*END``
        command, are written to the specified file without being
        executed.  An existing file of the same name, if any, will be
        overwritten.  Parameter values are not substituted for
        parameter names in the commands when the commands are written
        to the file.  Use ``*CFWRITE`` to create a file if this is
        desired.  The resulting macro may be executed with a ``*USE``
        command (which also allows parameters to be passed into the
        macro) or a /INPUT command (which does not allow parameters to
        be passed in).  Several macros may be stacked into a library
        file ``[*ULIB]``. You cannot use ``*CREATE`` within a DO loop.

        This command is valid in any processor.
        """
        return self.run(f"*CREATE,{fname},{ext}", **kwargs)

    def dflab(self, dof="", displab="", forcelab="", **kwargs):
        """Changes degree-of-freedom labels for user custom elements.

        APDL Command: /DFLAB

        Parameters
        ----------
        dof
            Number between 1 and 32 indicating which degree of freedom is to
            have its labels changed. For a list of these quantities, see the
            degree-of-freedom table in the echprm.inc file. The first few
            quantities follow:

        displab
            New label (four-character maximum) for the displacement label. The
            prior label is no longer valid.

        forcelab
            New label (four-character maximum) for the force label for this
            degree of freedom. The prior label is no longer valid.

        Notes
        -----
        The /DFLAB command is rarely used. Use it if you are writing a custom
        element and want to use degrees of freedom that are not part of the
        standard element set.
        """
        command = f"/DFLAB,{dof},{displab},{forcelab}"
        return self.run(command, **kwargs)

    def end(self, **kwargs):
        """Closes a macro file.

        APDL Command: ``*END``

        Notes
        -----
        Closes a file opened with ``*CREATE``. The ``*END`` command is an 8-character
        command (to differentiate it from ``*ENDIF``).   If you add commented text
        on that same line but do not allow enough spaces between ``*END`` and the
        "!" that indicates the comment text, the ``*END`` will attempt to interpret
        the "!" as the 8th character and will fail.

        This command is valid in any processor.
        """
        command = f"*END,"
        return self.run(command, **kwargs)

    def mkdir(self, dir_="", **kwargs):
        """Creates a directory.

        APDL Command: /MKDIR

        Parameters
        ----------
        dir_
            The directory to create (248 characters maximum on Linux;
            233 on Windows). If no path is provided, it will be
            created in the current working directory. Must be a valid
            name (and path) for the system you are working on.

        Notes
        -----
        It is recommended to just use ``os.mkdir``

        Creates a directory on the computer ANSYS is currently running on.
        """
        command = f"/MKDIR,{dir_}"
        return self.run(command, **kwargs)

    def msg(
        self,
        lab="",
        val1="",
        val2="",
        val3="",
        val4="",
        val5="",
        val6="",
        val7="",
        val8="",
        **kwargs,
    ):
        """Writes an output message via the ANSYS message subroutine.

        APDL Command: ``*MSG``

        Parameters
        ----------
        lab
            Label for output and termination control:

            Writes the message with no heading (default). - Writes the
            message with a "NOTE" heading.

            Writes the message with a "WARNING" heading.  Also writes
            the message to the errors file, Jobname.ERR. - Writes the
            message with a "ERROR" heading and causes run termination
            (if batch) at earliest "clean exit" point.  Also writes
            the message to the errors file, Jobname.ERR.

            Writes the message with a "FATAL ERROR" heading and causes
            run termination immediately.  Also writes the message to
            the errors file, Jobname.ERR. - Writes the message with a
            "NOTE" heading and displays it in the message dialog box.
            This option is most useful in GUI mode.

        val1, val2, val3, . . . , val8
            Numeric or alphanumeric character values to be included in message.
            Values may be the results of parameter evaluations.  All numeric
            values are assumed to be double precision. The FORTRAN nearest
            integer (NINT) function is used to form integers for the %I
            specifier.

        Notes
        -----
        Allows writing an output message via the ANSYS message subroutine.
        Also allows run termination control.  This command is used only when
        contained in a prepared file read into the ANSYS program (i.e.,
        ``*USE,/INPUT``, etc.).  A message format must immediately follow the ``*MSG``
        command (on a separate line, without parentheses, as described below).

        The message format may be up to 80 characters long, consisting of text
        strings and predefined "data descriptors" between the strings where
        numeric or alphanumeric character data are to be inserted.  The normal
        descriptors are %I for integer data, %G for double precision data, %C
        for alphanumeric character data, and %/ for a line break.  The
        corresponding FORTRAN data descriptors are I9, 1PG16.9 and A8,
        respectively.  Each descriptor must be preceded by a blank.  There must
        be one data descriptor for each specified value (8 maximum) in the
        order of the specified values.

        Enhanced descriptions may also be used:

        Do not begin ``*MSG`` format lines with ``*IF``, ``*ELSE`` ,
        ``*ELSEIF``, or ``*ENDIF`` .  If the last nonblank character
        of the message format is an ampersand (&), a second line will
        also be read as a continuation of the format.  Up to nine
        continuations (ten total lines) may be read.  If normal
        descriptions are used, then consecutive blanks are condensed
        into one blank upon output, and a period is appended.  Up to
        ten lines of output of 72 characters each may be produced
        (using the %/ descriptor).  Two examples follow.

        Here is an example of the ``*MSG`` command and a format to print a message
        with two integer values and one real value:

        The output line is:

        Here is an example illustrating multiline displays in GUI message
        windows:

        Note:: : The /UIS,MSGPOP command controls which messages are displayed
        in the message dialog box when the GUI is active.  All messages
        produced by the ``*MSG`` command are subject to the /UIS specification,
        with one exception,  If Lab = UI, the message will be displayed in the
        dialog box regardless of the /UIS specification.

        This command is valid in any processor.

        """
        command = f"*MSG,{lab},{val1},{val2},{val3},{val4},{val5},{val6},{val7},{val8}"
        return self.run(command, **kwargs)

    def pmacro(self, **kwargs):
        """Specifies that macro contents be written to the session log file.

        APDL Command: /PMACRO

        Notes
        -----
        This command forces the contents of a macro or other input file to be
        written to Jobname.LOG.  It is valid only within a macro or input file,
        and should be placed at the top of the file.  /PMACRO should be
        included in any macro or input file that calls GUI functions.
        """
        command = f"/PMACRO,"
        return self.run(command, **kwargs)

    def psearch(self, pname="", **kwargs):
        """Specifies a directory to be searched for "unknown command" macro files.

        APDL Command: /PSEARCH

        Parameters
        ----------
        pname
            Path name (64 characters maximum, and must include the final
            delimiter) of the middle directory to be searched.  Defaults to the
            user home directory.  If Pname = OFF, search only the ANSYS and
            current working directories.  If Pname = STAT, list the current
            middle directory and show the ANSYS_MACROLIB setting.

        Notes
        -----
        Specifies the pathname of a directory for file searches when reading
        "unknown command" macro files.  The search for the files is typically
        from the ANSYS directory, then from the user home directory, and then
        from the current working directory.  This command allows the middle
        directory searched to be other than the user home directory.

        This command is valid only at the Begin Level.
        """
        command = f"/PSEARCH,{pname}"
        return self.run(command, **kwargs)

    def rmdir(self, dir_="", **kwargs):
        """Removes (deletes) a directory.

        APDL Command: /RMDIR

        Parameters
        ----------
        dir\\_
            The directory to remove. If no path is provided, it will be assumed
            to be in the current working directory. All files in the directory
            are also removed.

        Notes
        -----
        Removes a directory on the computer ANSYS is currently running on. No
        warning or prompt is given, so use with extreme caution.
        """
        command = f"/RMDIR,{dir_}"
        return self.run(command, **kwargs)

    def tee(self, label="", fname="", ext="", **kwargs):
        """Writes a list of commands to a specified file at the same time that the

        APDL Command: /TEE
        commands are being executed.

        Parameters
        ----------
        label
            Indicates how ANSYS is to interpret this /TEE command:

            Signals the beginning of the command text that is to be
            written to Fname. If Fname already exists, specifying NEW
            causes the contents of Fname to be overwritten. -
            Indicates that you want to append to Fname the command
            text that follows.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        You can use the /TEE command to record a macro to a specified file at
        the same time that the macro is being executed. It is similar to the
        Linux tee command.

        For more information about the /TEE command, see the Introducing APDL
        of the ANSYS Parametric Design Language Guide.

        The following example illustrates the use of the /TEE command. If you
        issue these commands:

        the content of myfile.mac is:

        This command is valid in any processor, but only during an interactive
        run.
        """
        command = f"/TEE,{label},{fname},{ext}"
        return self.run(command, **kwargs)

    def ulib(self, fname="", ext="", **kwargs):
        """Identifies a macro library file.

        APDL Command: ``*ULIB``

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Identifies a macro library file for the ``*USE`` command.  A
        library of macros allows blocks of often used ANSYS commands
        to be stacked and executed from a single file.  The macro
        blocks must be enclosed within block identifier and terminator
        lines as shown in the example below.  If you want to add
        comment lines to a macro block, you may place them anywhere
        within the macro block.  (This includes placing them directly
        on the lines where the macro block identifier and the macro
        block terminator appear, as shown in the example.)  Do not
        place comment lines (or any other lines) outside of a macro
        block.

        The name of the macro library file is identified for reading
        on the ``*ULIB`` command.  The name of the macro block is
        identified on the ``*USE`` command.  The commands within the macro
        block are copied to a temporary file (of the macro block name)
        during the ``*USE`` operation and executed as if a macro file of
        that name had been created by the user.  The temporary file is
        deleted after it has been used.  Macro block names should be
        acceptable filenames (system dependent) and should not match
        user created macro file names, since the user macro file will
        be used first (if it exists) before the library file is
        searched.  Macro blocks may be stacked in any order.
        Branching [``*GO`` or ``*IF``] external to the macro block is not
        allowed.

        This command is valid in any processor.
        """
        command = f"*ULIB,{fname},{ext}"
        return self.run(command, **kwargs)

    def use(
        self,
        name="",
        arg1="",
        arg2="",
        arg3="",
        arg4="",
        arg5="",
        arg6="",
        arg7="",
        arg8="",
        arg9="",
        ar10="",
        ar11="",
        ar12="",
        ar13="",
        ar14="",
        ag15="",
        ar16="",
        ar17="",
        ar18="",
        **kwargs,
    ):
        """Executes a macro file.

        APDL Command: ``*USE``

        Parameters
        ----------
        name
            Name (32 characters maximum, beginning with a letter) identifying
            the macro file or a macro block on a macro library file.

        arg1, arg2, arg3, . . . , ar18
            Values passed into the file or block where the parameters ARG1
            through ARG9 and AR10 through AR18 are referenced.  Values may be
            numbers, alphanumeric character strings (up to 32 characters
            enclosed in single quotes), parameters (numeric or character) or
            parametric expressions.  See below for additional details.

        Notes
        -----
        Causes execution of a macro file called Name, or, if not found, a macro
        block "Name" on the macro library file [``*ULIB``].  Argument values
        (numeric or character) are passed into the file or block and
        substituted for local parameters ARG1, ARG2, ..., AR18.  The file Name
        may also be executed as an "unknown command" (i.e., without the ``*USE``
        command name) as described below.

        A macro is a sequence of ANSYS commands (as many as needed) recorded in
        a file or in a macro block in a library file (specified with the ``*ULIB``
        command).  The file or block is typically executed with the ``*USE``
        command.  In addition to command, numerical and alphanumeric data, the
        macro may include parameters which will be assigned numerical or
        alphanumerical character values when the macro is used.  Use of the
        macro may be repeated (within a do-loop, for example) with the
        parameters incremented.  A macro is defined within a run by "enclosing"
        a sequence of data input commands between a ``*CREATE`` and a ``*END``
        command.  The data input commands are passive (not executed) while
        being written to the macro file.  The macro file (without ``*CREATE`` and
        ``*END`` ) can also be created external to ANSYS.

        Up to 99 specially named scalar parameters called ARG1 to AR99 are
        locally available to each macro.  Note that the prefix for the first 9
        parameters is "ARG," while the prefix for the last 90 is "AR."  A local
        parameter is one which is not affected by, nor does it affect, other
        parameters, even those of the same name, which are used outside of the
        macro.  The only way a local parameter can affect, or be affected by,
        parameters outside the macro is if values are passed out of, or into,
        the macro by an argument list.  Parameters ARG1 through AR18 can have
        their values (numeric or character) passed via the argument list on the
        ``*USE`` command (ARG1 through AR19 can be passed as arguments on the
        "unknown command" macro).  Parameters AR19 through AR99 (AR20 through
        AR99 in the "unknown command" macro) are available solely for use
        within the macro; they cannot be passed via an argument list.  Local
        parameters are available to do-loops and to /INPUT files processed
        within the macro.  In addition to an ARG1--AR99 set for each macro,
        another ARG1--AR99 set is available external to all macros, local to
        "non-macro" space.

        A macro is exited after its last line is executed.  Macros may be
        nested (such as a ``*USE`` or an "unknown command" within a macro).  Each
        nested macro has its own set of 99 local parameters.  Only one set of
        local parameters can be active at a time and that is the set
        corresponding to the macro currently being executed or to the set
        external to all macros (if any).  When a nested macro completes
        execution, the previous set of local parameters once again becomes
        available.  Use ``*STATUS,ARGX`` to view current macro parameter values.

        An alternate way of executing a macro file is via the "unknown command"
        route.  If a command unknown to the ANSYS program is entered, a search
        for a file of that name (plus a .MAC suffix) is made.  If the file
        exists, it is executed, if not, the "unknown command" message is
        output.  Thus, users can write their own commands in terms of other
        ANSYS commands.  The procedure is similar to issuing the ``*USE`` command
        with the unknown command in the Name field.  For example, the command
        CMD,10,20,30 is internally similar to ``*USE,CMD,10,20,30``.  The macro
        file named CMD.MAC will be executed with the three parameters.  The
        ``*USE`` macro description also applies to the "unknown command" macro,
        except that various directories are searched and a suffix (.MAC) is
        assumed.  Also, a macro library file is not searched.

        A three-level directory search for the "unknown command" macro file may
        be available (see the Operations Guide).  The search order may be: 1) a
        high-level system directory, 2) the login directory, and 3) the local
        (working) directory.  Use the /PSEARCH command to change the directory
        search path.  For an "unknown command" CMD, the first file named
        CMD.MAC found to exist in the search order will be executed.  The
        command may be input as upper or lower case, however, it is converted
        to upper case before the file name search occurs.  On systems that
        uniquely support both upper and lower case file names, the file with
        the matching lower case name will be used if it exists, otherwise, the
        file with the matching upper case name will be used. All macro files
        placed in the apdl directory must be upper case.

        Note, since undocumented commands exist in the ANSYS program, the user
        should issue the command intended for the macro file name to be sure
        the "unknown command" message is output in the processor where it's to
        be used.  If the macro is to be used in other processors, the other
        processors must also be checked.

        This command is valid in any processor.
        """
        command = f"*USE,{name},{arg1},{arg2},{arg3},{arg4},{arg5},{arg6},{arg7},{arg8},{arg9},{ar10},{ar11},{ar12},{ar13},{ar14},{ag15},{ar16},{ar17},{ar18}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Generation:
    def rmalist(self, **kwargs):
        """Lists all defined master nodes for a ROM method.

        APDL Command: RMALIST

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMALIST,"
        return self.run(command, **kwargs)

    def rmanl(self, fname="", ext="", dimn="", oper="", **kwargs):
        """Assigns model database, dimensionality, and operating

        APDL Command: RMANL
        direction for the ROM method.

        Parameters
        ----------
        fname
            Database file name and directory path  (248 characters maximum,
            including directory). The file name defaults to Jobname.

        ext
            File extension (8 character maximum). The extension defaults to db.

        dimn
            Model dimensionality:

            2 - 2-D models

            3 - 3-D Models

        oper
            Primary operating direction:

            X - direction

            Y - direction

            Z - direction

        Notes
        -----
        Required Inputs:

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        return self.run(f"RMANL,{fname},{ext},,{dimn},{oper}", **kwargs)

    def rmaster(self, node="", lab="", **kwargs):
        """Defines master nodes for the ROM method.

        APDL Command: RMASTER

        Parameters
        ----------
        node
            Node number at which master degree of freedom is defined  If Node =
            P, graphical picking is enabled and all remaining command fields
            are ignored (valid only in the GUI).

        lab
            Valid labels are "ADD" (default) and "DEL".

        Notes
        -----
        Defines master nodes for the ROM.  Master nodes are used to track the
        total displacement of a structure in the operating direction [RMANL].
        They may be used as attachment points for 1-D structural elements
        during a ROM use pass via the UX degree of freedom.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMASTER,{node},{lab}"
        return self.run(command, **kwargs)

    def rmcap(self, refname="", c1="", c2="", **kwargs):
        """Defines lumped capacitance pairs between conductors C1 and C2 for a ROM

        APDL Command: RMCAP
        method.

        Parameters
        ----------
        refname
            Reference name for capacitance pair definition.

        c1
            First conductor (between 1 and 5).

        c2
            Second conductor (between 1 and 5).

        Notes
        -----
        For a capacitance definition between conductor C1 and C2, node
        components COND%C1% and COND%C2% (see CM command) must be present
        containing the conductor nodes. If C1 and C2 are blank, the capacitance
        definition with RefName will be deleted.  (For example, if C1 = 1, and
        C2 = 2, then node components COND1 and COND2 must be defined).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMCAP,{refname},{c1},{c2}"
        return self.run(command, **kwargs)

    def rmclist(self, **kwargs):
        """Lists all lumped capacitance pairs defined.

        APDL Command: RMCLIST

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMCLIST,"
        return self.run(command, **kwargs)

    def rmmlist(self, **kwargs):
        """Lists all mode specifications for the ROM method.

        APDL Command: RMMLIST

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMMLIST,"
        return self.run(command, **kwargs)

    def rmmrange(
        self,
        mode="",
        key="",
        min_="",
        max_="",
        nstep="",
        damp="",
        scale="",
        **kwargs,
    ):
        """Defines and edits various modal parameters for the ROM method.

        APDL Command: RMMRANGE

        Parameters
        ----------
        mode
            Mode number. Must be lower or equal to the number of modes
            extracted via the RMNEVEC command.

        key
            Mode classification key. Valid keys are:

            DOMINANT - Dominant mode

        min\\_
            Lower bound for fit range of mode.

        max\\_
            Upper bound for fit range of mode.

        nstep
            Number of equidistant steps in fit range of mode.

        damp
            Modal damping factor. Defaults to 0.0.

        scale
            Modal scaling factor.

        Notes
        -----
        When selected manually (RMMSELECT), modes must be classified as
        dominant, relevant, or unused. Dominant modes (Key = DOMINANT) are
        basis functions with large amplitudes. Relevant modes (Key = RELEVANT)
        are influenced by the dominant modes but do not cause interactions
        among themselves due to the small amplitude. This assumption leads to
        essential speed up of the sample point generator (see RMSMPLE).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMMRANGE,{mode},{key},{min_},{max_},{nstep},{damp},{scale}"
        return self.run(command, **kwargs)

    def rmmselect(self, nmode="", method="", dmin="", dmax="", **kwargs):
        """Selects modes for the ROM method.

        APDL Command: RMMSELECT

        Parameters
        ----------
        nmode
            Total number of modes to be selected

        method
            Method for mode selection. Valid methods are:

            TMOD - Automated selection using a test load. TMOD must be enclosed in single quotes.

        dmin
            Lower bound for total deflection range.

        dmax
            Upper bound for total deflection range.

        Notes
        -----
        Select pertinent modes for use in a ROM. Pertinent mode selection may
        be enhanced by using the deflection state of the structure
        representative of the operating nature of the device (Method = TMOD). A
        static analysis with an applied Test Load may be used.  The test load
        displacements must be extracted at the neutral plane of the device (if
        the device is stress-stiffened), or at any plane of the device (non-
        stress-stiffened). A node component "NEUN" must be defined for the
        plane of nodes, and the displacements extracted using the RMNDISP
        command prior to issuing this command. If Method = NMOD, use the first
        Nmode eigenmodes to select the pertinent modes for the ROM tool. Only
        those modes are selected that act in the operating direction of the
        structure [RMANL].

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.

        jobname.evx, jobname.evy, jobname.evz, jobname.evn, jobname.evl

        Test load and element load neutral plane displacement files:
        jobname.tld, jobname.eld
        """
        command = f"RMMSELECT,{nmode},{method},{dmin},{dmax}"
        return self.run(command, **kwargs)

    def rmporder(
        self,
        ord1="",
        ord2="",
        ord3="",
        ord4="",
        ord5="",
        ord6="",
        ord7="",
        ord8="",
        ord9="",
        **kwargs,
    ):
        """Defines polynomial orders for ROM functions.

        APDL Command: RMPORDER

        Parameters
        ----------
        ord1, ord2, ord3, . . . , ord9
            Polynomial orders for modes. Ordi specifies the polynomial order
            for modei. Modes are ordered as extracted from a modal analysis
            using the RMNEVEC command.  Defaults to 0 if mode i is unused;
            default to nstep(i) -1 for dominant or relevant modes, where
            nstep(i) is the number of equidistant steps in fit range of mode i.
            nstep(i) is automatically set by RMMSELECT or modified by the
            RMMRANGE command.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = (
            f"RMPORDER,{ord1},{ord2},{ord3},{ord4},{ord5},{ord6},{ord7},{ord8},{ord9}"
        )
        return self.run(command, **kwargs)

    def rmrgenerate(self, **kwargs):
        """Performs fitting procedure for all ROM functions to generate response

        APDL Command: RMRGENERATE
        surfaces.

        Notes
        -----
        The fitting procedure uses modal analysis data and function data
        generated using the RMSMPLE command and specifications set forth in the
        RMROPTIONS command. The files jobname_ijk.pcs (modes i, j, k) will be
        generated containing the coefficients of the response surfaces. These
        files are needed for the ROM Use Pass along with a ROM data base file
        [RMSAVE].

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.

        Strain energy and capacitance data file jobname_ijk.dec

        Response surface coefficients jobname_ijk.pcs (modes i, j, k)
        """
        command = f"RMRGENERATE,"
        return self.run(command, **kwargs)

    def rmroptions(self, refname="", type_="", invert="", **kwargs):
        """Defines options for ROM response surface fitting.

        APDL Command: RMROPTIONS

        Parameters
        ----------
        refname
            Reference name of ROM function to be fitted. Valid reference names
            are "SENE" for the strain energy of the structural domain and any
            capacitance reference name previously defined by means of RMCAP
            command for the electrostatic domain.

        type\\_
            Type of fitting function to be applied for regression analysis.
            Valid types are:

            LAGRANGE - Lagrange type (default)

        invert
            Flag to specify whether data should be inverted prior to fitting.

            0 - Do not invert data (default for SENE)

        Notes
        -----
        The objective of response surface fit is to compute an analytical
        expression for the strain energy and the capacitance as functions of
        modal amplitudes.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMROPTIONS,{refname},{type_},{invert}"
        return self.run(command, **kwargs)

    def rmrplot(self, refname="", type_="", mode1="", mode2="", **kwargs):
        """Plots response surface of ROM function or its derivatives with respect

        APDL Command: RMRPLOT
        to the dominant mode(s).

        Parameters
        ----------
        refname
            Reference name of ROM function. Valid reference names are "SENE"
            for the strain energy of the mechanical domain and any capacitance
            definition, previously defined by means of the RMCAP command, for
            the electrostatic domain.

        type\\_
            Type of data to be plotted. Valid types are:

            FUNC - Response surface (default)

        mode1
            First mode number (used for Type = "FIRST" and Type = "SECOND"
            only).

        mode2
            Second mode number (used for Type = "SECOND" only).

        Notes
        -----
        The objective of response surface fit is to compute an analytical
        expression for the strain energy and the capacitance as functions of
        modal amplitudes. This command assumes that the coefficient files
        jobnam_ijk.pcs are available [RMRGENERATE]. Visualization of the
        response surface will help to evaluate the validity of the function
        fit.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMRPLOT,{refname},{type_},{mode1},{mode2}"
        return self.run(command, **kwargs)

    def rmrstatus(self, refname="", **kwargs):
        """Prints status of response surface for ROM function.

        APDL Command: RMRSTATUS

        Parameters
        ----------
        refname
            Reference name of ROM function. Valid reference names are "SENE"
            for the strain energy of the mechanical domain and any capacitance
            reference names [RMCAP], for the electrostatic domain.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMRSTATUS,{refname}"
        return self.run(command, **kwargs)

    def rmsmple(self, nlgeom="", cap="", seqslv="", eeqslv="", **kwargs):
        """Runs finite element solutions and obtains sample points for the ROM

        APDL Command: RMSMPLE
        method.

        Parameters
        ----------
        nlgeom
            Specify whether a large or small deflection analysis is to be
            performed for the  mechanical domain:

            OFF (or 0) - Perform small deflection analysis (default).

        cap
            Capacitance calculation method.

            CHARGE - Compute capacitance based on the charge voltage relationship (default).

        seqslv
            Solver for structural analysis:

            SPARSE - Sparse direct equation solver (default).

        eeqslv
            Solver for electrostatic analysis:

            SPARSE - Sparse direct equation solver (default).

        Notes
        -----
        This command prepares and runs multiple finite element solutions on the
        Structural domain and the Electrostatic domain of a model to collect
        sample points of data for ROM response curve fitting. The command
        requires a model database [RMANL] and two Physics Files (Structural
        domain, titled "STRU" and an Electrostatic domain, titled "ELEC"; see
        PHYSICS command). Also required is a complete ROM database generated
        from the ROM Tools.  The Cap = CHARGE method is preferred when
        capacitance to "infinity" is not required. Capacitance conductor pairs
        are defined by the RMCAP command.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.

        Strain energy and capacitance data files  jobname_ijk.dec (mode i, j,
        k).
        """
        command = f"RMSMPLE,{nlgeom},{cap},{seqslv},{eeqslv}"
        return self.run(command, **kwargs)

    def rmxport(self, **kwargs):
        """Exports ROM model to external VHDL-AMS simulator.

        APDL Command: RMXPORT

        Notes
        -----
        Use this command to generate all files necessary to run the ROM
        analysis in an external VHDL-AMS Simulator.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.

        VHDL files: Initial.vhd, S_ams_ijk.vhd, Cxxx_ams_ijk.vhd,
        transducer.vhd.
        """
        command = f"RMXPORT,"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
These DATABASE commands can be used to initialize the database, save
it to a file, or annotate it with titles and systems of units.
"""


class Setup:
    def resume(self, fname="", ext="", nopar="", knoplot="", **kwargs):
        """Resumes the database from the database file.

        APDL Command: RESUME

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        nopar
            Parameter resume key:

            0
                All data in the database, including the scalar
                parameters, are replaced with the data saved on
                File.DB (default).

            1
                All data in the database, except the scalar
                parameters, are replaced with the data saved on
                File.DB.

        knoplot
            If equal to 1, will suppress automatic plot. Otherwise, if
            the GUI is on and this RESUME command was not read from a
            file, the selected elements from Fname are plotted. (If
            there are no selected elements, selected nodes are
            plotted. If no nodes, volumes; if no volumes, areas; if no
            areas, lines; if no lines, keypoints. If there are no
            selected keypoints, the screen is erased.)

        Notes
        -----
        The RESUME command resumes a database file into the ANSYS program. The
        command causes the database file (File.DB) to be read, thereby
        resetting the database (including any geometry settings) either a) as
        it was at the last SAVE command, or b) as it was saved with the last
        /EXIT command, whichever was last.

        For multiple load step analyses (because only the data for one load
        step at a time may reside in the database), the load step data restored
        to the database will correspond to the load step data written when the
        save occurred.

        If the database file was saved in another ANSYS, Inc. product, it may
        contain element type and KEYOPT specifications which are invalid in the
        resuming product. Immediately after the database resume is completed,
        you should redefine these invalid element types and KEYOPT settings to
        valid ones (ET, KEYOPT).

        The NOPAR = 1 option should not be used if array parameters are
        defined, as existing array parameters might be redefined with arbitrary
        values. For a more general method of preventing the replacement of both
        scalar and array parameters, see PARSAV and PARRES.)

        This command is valid in any processor.  If used in the solution
        processor, this command is valid only within the first load step.
        """
        return self.run(f"RESUME,{fname},{ext},,{nopar},{knoplot}", **kwargs)

    def save(self, fname="", ext="", slab="", **kwargs):
        """Saves all current database information.

        APDL Command: SAVE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        slab
            Mode for saving the database:

            ALL
              Save the model data, solution data and post data
              (element tables, etc.). This value is the default.

            MODEL
              Save the model data (solid model, finite element
              model, loadings, etc.) only.

            SOLU
              Save the model data and the solution data (nodal
              and element results).

        Notes
        -----
        Saves all current database information to a file (File.DB).
        In interactive mode, an existing File.DB is first written to a
        backup file (File.DBB).  In batch mode, an existing File.DB is
        replaced by the current database information with no backup.
        The command should be issued periodically to ensure a current
        file backup in case of a system "crash" or a "line drop."  It
        may also be issued before a "doubtful" command so that if the
        result is not what was intended the database may be easily
        restored to the previous state.  A save may be time consuming
        for large models.  Repeated use of this command overwrites the
        previous data on the file (but a backup file is first written
        during an interactive run).  When issued from within POST1,
        the nodal boundary conditions in the database (which were read
        from the results file) will overwrite the nodal boundary
        conditions existing on the database file.

        Internal nodes may be created during solution (for example,
        via the mixed u-P formulation or generalized plane strain
        option for current- technology elements, the Lagrangian
        multiplier method for contact elements or the MPC184 elements,
        or the quadratic or cubic option of the BEAM188 and PIPE288
        elements). It is sometimes necessary to save the internal
        nodes in the database for later operations, such as cutting
        boundary interpolations (CBDOF) for submodeling. To do so,
        issue the SAVE command after the first SOLVE command.

        In general, saving after solving is always a good practice.

        This command is valid in any processor.
        """
        return self.run(f"SAVE,{fname},{ext},,{slab}", **kwargs)

    def smbc(self, mode="", **kwargs):
        """Controls the display of solid model boundary condition symbols and

        APDL Command: /SMBC
        labels.

        Parameters
        ----------
        mode
            CENT
              Solid model boundary condition symbols and labels appear at the centroid of the
              solid model entity (default).

            TESS
              Solid model boundary condition symbols and labels appear inside each
              constituent element of the tessellation.

        Notes
        -----
        Mode = CENT is designed to reduce the clutter of boundary condition
        symbols in solid model plots. For example, if you have assigned normal
        pressure loads to an area, you may choose to display the pressures as
        arrows with the /PSF command using /PSF,PRES,NORM,2. When Mode = CENT,
        the pressure arrow is displayed at the centroid of the area. When Mode
        = TESS, a pressure arrow is displayed at the centroid of each polygon
        of the area's tessellation.

        This command is valid in any processor.
        """
        command = "/SMBC,%s" % (str(mode))
        return self.run(command, **kwargs)

    def stat(self, **kwargs):
        """Displays the status of database settings.

        APDL Command: STAT

        Notes
        -----
        In the DISPLAY program, STAT will show the current status of the
        program settings.

        In the ANSYS program, STAT is a command generated by the GUI and will
        appear in the log file (Jobname.LOG) if status is requested for some
        items under Utility Menu> List> Status.  Generally, STAT will be
        preceded by one of the commands listed below, which specifies the
        particular topic that status was requested for.

        If entered directly into the program, the STAT command should be
        immediately preceded by the desired topic command listed below.  In
        processors other than those listed below (e.g., AUX12), no topic
        command should proceed STAT.

        This command is valid in any processor.

        PREP7 topic commands (and their corresponding topics) are:

        SOLUTION topic commands (and their corresponding topics) are:

        POST1 topic commands (and their corresponding topics) are:

        POST26 topic commands (and their corresponding topics) are:
        """
        command = "STAT,"
        return self.run(command, **kwargs)

    def stitle(self, nline="", title="", **kwargs):
        """Defines subtitles.

        APDL Command: /STITLE

        Parameters
        ----------
        nline
            Subtitle line number (1 to 4).  Defaults to 1.

        title
            Input up to 70 alphanumeric characters.  Parameter substitution may
            be forced within the title by enclosing the parameter name or
            parametric expression within percent (%) signs.  If Title is blank,
            this subtitle is deleted.

        Notes
        -----
        Subtitles (4 maximum) are displayed in the output along with the main
        title [/TITLE].  Subtitles do not appear in GUI windows or in ANSYS
        plot displays.  The first subtitle is also written to various ANSYS
        files along with the main title.  Previous subtitles may be overwritten
        or deleted.  Issue /STATUS to display titles.

        This command is valid in any processor.
        """
        command = "/STITLE,%s,%s" % (str(nline), str(title))
        return self.run(command, **kwargs)

    def title(self, title="", **kwargs):
        """Defines a main title.

        APDL Command: /TITLE

        Parameters
        ----------
        title str
            Input up to 72 alphanumeric characters.  Parameter substitution may
            be forced within the title by enclosing the parameter name or
            parametric expression within percent (%) signs.

        Notes
        -----
        The title is carried through the printout and written on various files.
        The title written to a file is the title defined at that time.  Special
        characters may be used within the title text.  Subtitles may also be
        defined [/STITLE].

        This command is valid in any processor.
        """
        return self.run(f"/TITLE,{title}", **kwargs)

    def units(
        self,
        label="",
        lenfact="",
        massfact="",
        timefact="",
        tempfact="",
        toffset="",
        chargefact="",
        forcefact="",
        heatfact="",
        **kwargs,
    ):
        """Annotates the database with the system of units used.

        APDL Command: /UNITS

        Parameters
        ----------
        label
            Label to denote the system of units used in this job:

            USER
                User-defined system (default).

            SI
                International system (m, kg, s, K).

            MKS
                MKS system (m, kg, s, °C).

            uMKS
                μMKS system (μm, kg, s, °C).

            CGS
                CGS system (cm, g, s, °C).

            MPA
                MPA system (mm, Mg, s, °C).

            BFT
                U. S. Customary system using feet (ft, slug, s, °F).

            BIN
                U. S. Customary system using inches (in, lbf*s2/in, s, °F).

        Notes
        -----
        Allows the user to set a marker in the database indicating the system
        of units used.  The setting may be reviewed with the /STATUS command at
        the Begin level.  The units label and conversion factors on this
        command are for user convenience only and have no effect on the
        analysis or data.  That is, /UNITS will not convert database items from
        one system to another (e.g., from U. S. Customary to SI, etc.).  The
        units setting will be written to the file of IGES data [IGESOUT or
        CDWRITE], which can then be read by many programs that read IGES files.
        The user must still use consistent units for the results to be valid.

        If you choose the MKS system of units, the EPZRO option for the EMUNIT
        command is set to 8.85 e-12 F/m.  (EPZRO specifies alternate free-space
        permittivity.)

        For micro-electromechanical systems (MEMS), where dimensions are on the
        order of microns, see the conversion factors in System of Units in the
        Coupled-Field Analysis Guide.

        If you use the ANSYS ADAMS Interface to export model information to the
        ADAMS program, the /UNITS command is required to ensure the correct
        transfer of data between ANSYS and ADAMS. You may choose a predefined
        unit system label (Label = SI, CGS, etc.) or you can select the user-
        defined system option (Label = USER) and input the appropriate
        conversion factors (LENFACT, MASSFACT, TIMEFACT, and FORCEFACT). The
        conversion factors will be written to the ADAMS input file Jobname.MNF
        in order to correctly generate the load. For more information, see
        Export to ADAMS in the Substructuring Analysis Guide.

        All differences between the base solution units used by the ANSYS and
        CFX solvers will be noted in the ANSYS output file.   Unit conversions
        are automatically applied to all loads transferred unless Label = USER.
        Unit conversions are not applied to any of the loads transferred
        between the ANSYS and CFX solvers if they use a user-defined unit
        system.

        This command is valid in any processor.
        """
        command = "/UNITS,%s,%s,%s,%s,%s,%s,%s,%s,%s" % (
            str(label),
            str(lenfact),
            str(massfact),
            str(timefact),
            str(tempfact),
            str(toffset),
            str(chargefact),
            str(forcefact),
            str(heatfact),
        )
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from typing import Dict, Union

import click


@click.command(
    short_help="Launch MAPDL instances.",
    help="""This command aims to replicate the behavior of :func:`ansys.mapdl.core.launcher.launch_mapdl`

For more information see :func:`ansys.mapdl.core.launcher.launch_mapdl`.""",
)
@click.option(
    "--exec_file",
    default=None,
    type=str,
    help="The location of the MAPDL executable.  Will use the cached location when left at the default ``None`` and no environment variable is set. The executable path can be also set through the environment variable ``PYMAPDL_MAPDL_EXEC``.",
)
@click.option(
    "--run_location",
    default=None,
    type=str,
    help="MAPDL working directory.  Defaults to a temporary working directory.  If directory doesn't exist, one is created.",
)
@click.option(
    "--jobname",
    default="file",
    type=str,
    help="MAPDL jobname.  Defaults to ``'file'``.",
)
@click.option(
    "--nproc", type=int, default=2, help="Number of processors.  Defaults to 2."
)
@click.option(
    "--ram",
    default=None,
    type=int,
    help="Fixed amount of memory to request for MAPDL.  If ``None``, then MAPDL will use as much as available on the host machine.",
)
@click.option(
    "--mode",
    type=str,
    default=None,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--override",
    default=False,
    type=bool,
    help="Attempts to delete the lock file at the ``run_location``. Useful when a prior MAPDL session has exited prematurely and the lock file has not been deleted.",
)
@click.option(
    "--loglevel",
    default="",
    type=str,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--additional_switches",
    default="",
    type=str,
    help="Additional switches for MAPDL, for example ``'aa_r'``, the academic research license. Avoid adding switches like ``-i``, ``-o`` or ``-b`` as these are already included to start up the MAPDL server.",
)
@click.option(
    "--start_timeout",
    default=45,
    type=int,
    help="Maximum allowable time to connect to the MAPDL server.",
)
@click.option(
    "--port",
    default=None,
    type=int,
    help="Port to launch MAPDL gRPC on.  Final port will be the first port available after (or including) this port.  Defaults to 50052. You can also override the port default with the environment variable ``PYMAPDL_PORT=<VALID PORT>`` This argument has priority over the environment variable.",
)
@click.option(
    "--cleanup_on_exit",
    default=False,
    type=bool,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--start_instance",
    default=None,
    type=bool,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--ip",
    default=None,
    type=str,
    help="Argument not allowed in CLI. It will be ignored",
)
@click.option(
    "--clear_on_connect",
    default=False,
    type=bool,
    help="Argument not allowed in CLI. It will be ignored",
)
@click.option(
    "--log_apdl",
    default=None,
    type=str,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--remove_temp_dir_on_exit",
    default=False,
    type=bool,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--license_server_check",
    default=False,
    type=bool,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--license_type",
    default=None,
    type=str,
    help="Enable license type selection. You can input a string for its license name (for example ``'meba'`` or ``'ansys'``) or its description ('enterprise solver' or 'enterprise' respectively). You can also use legacy licenses (for example ``'aa_t_a'``) but it will also raise a warning. If it is not used (``None``), no specific license will be requested, being up to the license server to provide a specific license type. Default is ``None``.",
)
@click.option(
    "--print_com",
    default=False,
    type=bool,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--add_env_vars",
    default=None,
    type=str,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--replace_env_vars",
    default=None,
    type=str,
    help="Argument not allowed in CLI. It will be ignored.",
)
@click.option(
    "--version",
    default=None,
    type=str,
    help="Version of MAPDL to launch. If ``None``, the latest version is used. Versions can be provided as integers (i.e. ``version=222``) or floats (i.e. ``version=22.2``). To retrieve the available installed versions, use the function :meth:`ansys.tools.path.path.get_available_ansys_installations`.",
)
def start(
    exec_file: str,
    run_location: str,
    jobname: str,
    nproc: Union[int, str],
    ram: Union[int, str],
    mode: str,  # ignored
    override: bool,
    loglevel: str,  # ignored
    additional_switches: str,
    start_timeout: Union[int, str],
    port: Union[int, str],
    cleanup_on_exit: bool,  # ignored
    start_instance: bool,  # ignored
    ip: str,
    clear_on_connect: bool,  # ignored
    log_apdl: bool,  # ignored
    remove_temp_dir_on_exit: bool,  # ignored
    license_server_check: bool,  # ignored
    license_type: str,
    print_com: bool,  # ignored
    add_env_vars: Dict[str, str],  # ignored
    replace_env_vars: Dict[str, str],  # ignored
    version: Union[int, str],
):
    from ansys.mapdl.core.launcher import launch_mapdl

    if mode:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'mode'.\nIgnoring argument."
        )

    if loglevel:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'loglevel'.\nIgnoring argument."
        )

    if cleanup_on_exit:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'cleanup_on_exit'.\nIgnoring argument."
        )

    if start_instance:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'start_instance'.\nIgnoring argument."
        )

    if ip:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'ip'.\nIgnoring argument."
        )

    if clear_on_connect:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'clear_on_connect'.\nIgnoring argument."
        )

    if log_apdl:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'log_apdl'.\nIgnoring argument."
        )

    if remove_temp_dir_on_exit:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'remove_temp_dir_on_exit'.\nIgnoring argument."
        )

    if print_com:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'print_com'.\nIgnoring argument."
        )

    if add_env_vars:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'add_env_vars'.\nIgnoring argument."
        )

    if replace_env_vars:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'replace_env_vars'.\nIgnoring argument."
        )

    if license_server_check:
        click.echo(
            click.style("Warn:", fg="yellow")
            + " The following argument is not allowed in CLI: 'license_server_check'.\nIgnoring argument."
        )

    # Ignoring env var if using CLI
    if "PYMAPDL_START_INSTANCE" in os.environ:
        os.environ.pop("PYMAPDL_START_INSTANCE")

    out = launch_mapdl(
        exec_file=exec_file,
        just_launch=True,
        run_location=run_location,
        jobname=jobname,
        nproc=nproc,
        ram=ram,
        override=override,
        additional_switches=additional_switches,
        start_timeout=start_timeout,
        port=port,
        license_server_check=license_server_check,
        license_type=license_type,
        version=version,
    )

    if len(out) == 3:
        header = f"Launched an MAPDL instance (PID={out[2]}) at "
    else:
        header = "Launched an MAPDL instance at "

    click.echo(click.style("Success: ", fg="green") + header + f"{out[0]}:{out[1]}")



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class SolidForces:
    def fk(self, kpoi="", lab="", value="", value2="", **kwargs):
        """Defines force loads at keypoints.

        APDL Command: FK

        Parameters
        ----------
        kpoi
            Keypoint at which force is to be specified.  If ALL, apply to all
            selected keypoints [KSEL].  If KPOI = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be substituted for KPOI.

        lab
            Valid force label.  Structural labels:  FX, FY, or FZ (forces); MX,
            MY, or MZ (moments).  Thermal labels:  HEAT, HBOT, HE2, HE3, . . .,
            HTOP (heat flow).  Fluid labels:  FLOW (fluid flow).  Electric
            labels:  AMPS (current flow), CHRG (electric charge).  Magnetic
            labels:  FLUX (magnetic flux);  CSGX, CSGY, or CSGZ (magnetic
            current segments). Diffusion labels: RATE (diffusion flow rate).

        value
            Force value or table name reference for specifying tabular boundary
            conditions.  To specify a table, enclose the table name in percent
            signs (%), e.g., FK, KPOI, HEAT,%tabname%).  Use the ``*DIM`` command
            to define a table.

        value2
            Second force value (if any).  If the analysis type and the force
            allow a complex input, VALUE (above) is the real component and
            VALUE2 is the imaginary component.

        Notes
        -----
        Forces may be transferred from keypoints to nodes with the FTRAN or
        SBCTRAN commands.  See the F command for a description of force loads.

        Tabular boundary conditions (VALUE = %tabname%) are available only for
        the following labels: Fluid (FLOW), Electric (AMPS), Structural force
        (FX, FY, FZ, MX, MY, MZ), and Thermal (HEAT, HBOT, HE2, HE3, . . .,
        HTOP).

        This command is also valid in PREP7.
        """
        command = f"FK,{kpoi},{lab},{value},{value2}"
        return self.run(command, **kwargs)

    def fkdele(self, kpoi="", lab="", **kwargs):
        """Deletes force loads at a keypoint.

        APDL Command: FKDELE

        Parameters
        ----------
        kpoi
            Keypoint at which force is to be deleted.  If ALL, delete forces at
            all selected keypoints [KSEL].  If KPOI = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be substituted for KPOI.

        lab
            Valid force label.  If ALL, use all appropriate labels.  See the
            FDELE command for labels.

        Notes
        -----
        Deletes force loads (and all corresponding finite element loads) at a
        keypoint.  See the FDELE command for details.

        This command is also valid in PREP7.
        """
        command = f"FKDELE,{kpoi},{lab}"
        return self.run(command, **kwargs)

    def fklist(self, kpoi="", lab="", **kwargs):
        """Lists the forces at keypoints.

        APDL Command: FKLIST

        Parameters
        ----------
        kpoi
            List forces at this keypoint.  If ALL (default), list for all
            selected keypoints [KSEL].  If KPOI = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be substituted for KPOI.

        lab
            Force label to be listed (defaults to ALL).  See the DOFSEL command
            for labels.

        Notes
        -----
        Listing applies to the selected keypoints [KSEL] and the selected force
        labels [DOFSEL].

        This command is valid in any processor.
        """
        command = f"FKLIST,{kpoi},{lab}"
        return self.run(command, **kwargs)

    def ftran(self, **kwargs):
        """Transfers solid model forces to the finite element model.

        APDL Command: FTRAN

        Notes
        -----
        Forces are transferred only from selected keypoints to selected nodes.
        The FTRAN operation is also done if the SBCTRAN command is issued or
        automatically done upon initiation of the solution calculations
        [SOLVE].

        This command is also valid in PREP7.
        """
        command = f"FTRAN,"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Scaling:
    def slashdscale(self, wn="", dmult="", **kwargs):
        """Sets the displacement multiplier for displacement displays.

        APDL Command: /DSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        dmult
            AUTO or 0
                Scale displacements automatically so that maximum displacement
                (vector amplitude) displays as 5 percent of the maximum model
                length, as measured in the global Cartesian X, Y, or Z
                directions.

            1
                Do not scale displacements (that is, scale displacements by 1.0,
                true to geometry). Often used with large deflection results.

            FACTOR
                Scale displacements by numerical value input for `FACTOR`.

            OFF
                Remove displacement scaling (that is, scale displacements by 0.0,
                no distortion).

            USER
                Set `dmult` to that used for last display (useful when last `dmult`
                value was automatically calculated).

        Notes
        -----

        **Command Default**

        The default value is AUTO or 0 except when:

        * Large deflection effects are included (`NLGEOM,ON`) and it is not a
          modal analysis; then the default is 1.

        * It is a spectrum analysis (`ANTYPE,SPECTR`); then the default is `OFF`.

        * The amplitude of a time-harmonic solution is computed using the `HRCPLX`
          command (`OMEGAT ≥ 360°`); then the default is `OFF`.

        * The amplitude of a complex modal or harmonic solution is stored into
          the database using the `SET` command (`KIMG = AMPL`); then the default
          is `OFF`.

        If Multi-Plots are not being displayed, and the current device is a 3D
        device (`/SHOW,3D`), then the displacement scale in all active windows
        will be the same, even if separate `/DSCALE` commands are issued for
        each active window. For efficiency, the program maintains a single data
        structure (segment) containing only one displacement scale. The program
        displays the same segment (displacement scale) in all windows. Only the
        view settings will be different in each of the active windows.

        This command is valid in any processor.

        """
        command = f"/DSCALE, {wn}, {dmult}"
        return self.run(command, **kwargs)

    def iclwid(self, factor="", **kwargs):
        """Scales the line width of circuit builder icons.

        APDL Command: /ICLWID

        Parameters
        ----------
        factor
            Multiplication factor applied to the default line width (defaults
            to 1). The minimum is 1 and the maximum is 6.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"/ICLWID,{factor}"
        return self.run(command, **kwargs)

    def icscale(self, wn="", factor="", **kwargs):
        """Scales the icon size for elements supported in the circuit builder.

        APDL Command: /ICSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        factor
            Factor applied to the default icon size (defaults to 1).

        Notes
        -----
        Scaling the icon size can provide better visualization of the circuit
        components when using the Circuit Builder (an interactive builder
        available in the ANSYS GUI).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"/ICSCALE,{wn},{factor}"
        return self.run(command, **kwargs)

    def ratio(self, wn="", ratox="", ratoy="", **kwargs):
        """Distorts the object geometry.

        APDL Command: /RATIO

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        ratox
            Distort object in the window X direction by this factor (defaults
            to 1.0).

        ratoy
            Distort object in the window Y direction by this factor (defaults
            to 1.0).

        Notes
        -----
        Distorts the object geometry in a particular direction.  An example of
        this command's use would be to allow long narrow sections to be
        distorted to a more square area for better display visualization.

        This command is valid in any processor.
        """
        command = f"/RATIO,{wn},{ratox},{ratoy}"
        return self.run(command, **kwargs)

    def shrink(self, ratio="", **kwargs):
        """Shrinks elements, lines, areas, and volumes for display clarity.

        APDL Command: /SHRINK

        Parameters
        ----------
        ratio
            Shrinkage ratio (input as a decimal (0.0 to 0.5)).  Defaults to 0.0
            (no shrinkage).  Values greater than 0.5 default to 0.1 (10%
            shrinkage).

        Notes
        -----
        Shrinks the elements, lines, areas, and volumes so that adjacent
        entities are separated for clarity.  Portions of this command are not
        supported by PowerGraphics [/GRAPHICS,POWER].

        If only the common lines of non-coplanar faces are drawn (as per the
        /EDGE command), then this command is ignored.

        This command is valid in any processor.
        """
        command = f"/SHRINK,{ratio}"
        return self.run(command, **kwargs)

    def sscale(self, wn="", smult="", **kwargs):
        """Sets the contour multiplier for topographic displays.

        APDL Command: /SSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        smult
            Contour multiplier that factors in results based on the product of
            the multiplier and the result being plotted. Defaults to 0.0 (no
            topographic effects).

        Notes
        -----
        Use this command to scale values to the geometry when the contours are
        shown elevated.  For section displays [/TYPE], the elevation is
        performed perpendicular to the section face.

        Nonzero contour multipliers factoring in large results (stresses or
        displacements) can produce very large distortion, causing images to
        disappear.  To bring a distorted image back into view, reduce the
        contour multiplier value.

        Portions of this command are not supported by PowerGraphics
        [/GRAPHICS,POWER].
        """
        command = f"/SSCALE,{wn},{smult}"
        return self.run(command, **kwargs)

    def txtre(self, lab="", num="", n1="", n2="", ninc="", **kwargs):
        """Controls application of texture to selected items.

        APDL Command: /TXTRE

        Parameters
        ----------
        lab
            You can apply texture according to the following labels:

            ELEM
                Apply texture to elements N1 through N2 in steps of NINC.

            AREA
                Apply texture to areas N1 through N2 in steps of NINC.

            VOLU
                Apply texture to volumes N1 through N2 in steps of NINC.

            CM
                Apply texture to the component named in N1. N2 and
                NINC are ignored.

            ON, OFF
                Sets the specified texture display on or
                off. All other fields are ignored.

            File
                If Lab = File, the command format is /TXTRE, File,
                Key_Index, Fname, Fext, --, Format (This variant of
                the command is applicable to 2-D drivers).

            Key_Index
                The texture index associated with the file. If
                the number fifty-one (51) is used, the
                imported bitmap will be used as the window's
                logo.

            Fname
                File name and directory path (248 characters
                maximum, including the characters needed for the
                directory path).  An unspecified directory path
                defaults to the working directory; in this case,
                you can use all 248 characters for the file name.

            Fext
                Filename extension (eight-character maximum).

            Format
                The file format. If Format = 0, the file is a
                pixmap (Linux) or Bitmap (PC).  The file cannot
                contain a compressed image, and the PC file must
                be 8 or 24 bit BI_RGB format. If Format = 1 or
                JPEG, then the file is in JPEG (Joint
                Photographic Experts Group) format. If Format = 2
                or PNG, then the file is in PNG (Portable Network
                Graphics) format.

        num
            Select the texture index number from the following list:

            0
                No Texturing

            1
                Aluminum

            2
                 Aluminum, Brushed

            3
                Steel With Bumps

            4
                Steel, Embossed

            5
                Iron

            6
                Steel, Pattern

            7
                Steel, Riveted

            8
                Steel, Scratched

            9
                Tin

            10
                Metal

            11
                Steel, Etched

            12
                Metal, Hot

            13
                Iron, Grainy

            14
                Metal, Rusty

            15
                Brick

            16
                Block

            17
                Wood

            18
                Wood, Light

            19
                Wood, Walnut

            20
                Plastic, Hard Blue

            21
                Plastic, Light Blue

            22
                Plastic, Hard Red

            31
                Gold

            32
                Brass

            33
                Silver

            34
                 Plastic, Black

            35
                Plastic, Ivory

            36
                Plastic, Blue

            37
                Plastic, Red

            38
                Plastic, Yellow

            39
                Plastic, Green

            40
                Plastic, Brown

        n1, n2, ninc
            Apply texture to Lab items numbered N1 through N2 in steps of NINC
            (defaults to 1). If Lab = CM, then N1 is used to for the component
            name and N2 and NINC are ignored. If Lab = ELEM, AREA, or VOLU and
            N1 = blank or ALL, then the specified texture will be applied to
            all entities of type Lab. If N1 = P, then graphical picking is
            enabled.

        Notes
        -----
        This command is available for 3-D Open GL devices. 2-D devices are
        supported only for the Lab = File variation of the command, allowing
        imported bitmaps to be used for texturing and annotation. Textures can
        affect the speed of many of your display operations. You can increase
        the speed by temporarily turning the textures off (Utility Menu>
        PlotCtrls> Style> Texturing(3D)> Display Texturing). This menu
        selection toggles your textures on and off. When textures are toggled
        off, all of the texture information is retained and reapplied when
        texturing is toggled back on.

        For some displays, the texture will appear distorted because of a
        technique used to enhance 3-D displays (/DV3D,TRIS,1). Disabling this
        function (/DV3D,TRIS,0) will improve the quality of some texture
        displays. Disabling the TRIS option of the /DV3D command will slow down
        3-D displays significantly. Be sure to reapply the TRIS option after
        you obtain a satisfactory output.

        Specifying /TXTRE,DEFA removes all texturing.
        """
        command = f"/TXTRE,{lab},{num},{n1},{n2},{ninc}"
        return self.run(command, **kwargs)

    def vscale(self, wn="", vratio="", key="", **kwargs):
        """Scales the length of displayed vectors.

        APDL Command: /VSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        vratio
            Ratio value applied to the automatically calculated scale factor
            (defaults to 1.0, i.e., use scale factor as automatically
            calculated).

        key
            Relative scaling key:

            0
                Use relative length scaling among vectors based on magnitudes.

            1
                Use uniform length scaling for all vector lengths.

        Notes
        -----
        Allows scaling of the vector length displayed with the PLVECT command
        of POST1 and the /PBC and /PSF commands.  Also allows the scaling of
        the  element (i.e., /PSYMB,ESYS) and the nodal (i.e., /PSYMB,NDIR)
        coordinate system symbols.

        This command is valid in any processor.
        """
        command = f"/VSCALE,{wn},{vratio},{key}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.mapdl.core._commands import parse


class Lines:
    def bsplin(
        self,
        p1="",
        p2="",
        p3="",
        p4="",
        p5="",
        p6="",
        xv1="",
        yv1="",
        zv1="",
        xv6="",
        yv6="",
        zv6="",
        **kwargs,
    ) -> int:
        """Generate a single line from a spline fit to a series of keypoints.

        APDL Command: BSPLIN

        Parameters
        ----------
        p1, p2, p3, p4, p5, p6
            Keypoints through which a spline is fit.  At least two
            keypoints must be defined.

        XV1, YV1, ZV1
            Orientation point of an outward vector tangent to line at
            P1. Vector coordinate system has its origin at the
            keypoint. Coordinate interpretation corresponds to the
            active coordinate system type, i.e., X is R for
            cylindrical, etc. Defaults to zero curvature slope.

        XV6, YV6, ZV6
            Orientation point of an outward vector tangent to a line
            at P6 (or the last keypoint specified if fewer than six
            specified). Defaults to zero curvature slope.

        Returns
        -------
        int
            Line number of the spline generated from the spline fit.

        Notes
        -----
        One line is generated between keypoint P1 and the last
        keypoint entered.  The line will pass through each entered
        keypoint.  Solid modeling in a toroidal coordinate system is
        not recommended.

        Examples
        --------
        Generate a spline through ``(0, 0, 0)``, ``(0, 1, 0)`` and
        ``(1, 2, 0)``

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 1, 0)
        >>> k2 = mapdl.k("", 1, 2, 0)
        >>> lnum = mapdl.bsplin(k0, k1, k2)

        """
        command = (
            f"BSPLIN,{p1},{p2},{p3},{p4},{p5},{p6},{xv1},{yv1},{zv1},{xv6},{yv6},{zv6}"
        )
        return parse.parse_line_no(self.run(command, **kwargs))

    def circle(
        self, pcent="", rad="", paxis="", pzero="", arc="", nseg="", **kwargs
    ) -> list:
        """Generate circular arc lines.

        APDL Command: CIRCLE

        Generates circular arc lines (and their corresponding
        keypoints).  Keypoints are generated at regular angular
        locations (based on a maximum spacing of 90 degrees).  Arc lines are
        generated connecting the keypoints.  Keypoint and line numbers
        are automatically assigned, beginning with the lowest
        available values [NUMSTR].  Adjacent lines use a common
        keypoint.  Line shapes are generated as arcs, regardless of
        the active coordinate system.  Line shapes are invariant with
        coordinate system after they are generated.

        Parameters
        ----------
        pcent
            Keypoint defining the center of the circle (in the plane
            of the circle).

        rad
            Radius of the circle.  If RAD is blank and PCENT = P, the
            radius is the distance from PCENT to PZERO.

        paxis
            Keypoint defining axis of circle (along with PCENT).  If
            PCENT = P and PAXIS is omitted, the axis is normal to the
            working plane.

        pzero
            Keypoint defining the plane normal to circle (along with
            PCENT and PAXIS) and the zero degree location.  Need not
            be in the plane of the circle. This value is not required
            if PAXIS is defined along the Y axis (that is, a circle in
            the XZ plane).

        arc
            Arc length (in degrees).  Positive follows right-hand rule
            about PCENT-PAXIS vector.  Defaults to 360 degrees.

        nseg
            Number of lines around circumference (defaults to minimum
            required for 90 degrees-maximum arcs, i.e., 4 for 360 degrees).  Number
            of keypoints generated is NSEG for 360 degrees or NSEG + 1 for
            less than 360 degrees.

        Returns
        -------
        list
            List of lines of the circular arcs generated from this
            command.

        Examples
        --------
        Create a full circle containing four circular arcs.  Circle
        centered at (0, 0, 0) and generated in the XY plane.  Return
        the lines generated from the circle.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 0, 1)
        >>> carc0 = mapdl.circle(k0, 1, k1)
        >>> carc0
        [1, 2, 3, 4]

        """
        command = f"CIRCLE,{pcent},{rad},{paxis},{pzero},{arc},{nseg}"
        return parse.parse_line_nos(self.run(command, **kwargs))

    def l(
        self,
        p1="",
        p2="",
        ndiv="",
        space="",
        xv1="",
        yv1="",
        zv1="",
        xv2="",
        yv2="",
        zv2="",
        **kwargs,
    ) -> int:
        """Define a line between two keypoints.

        APDL Command: L

        Defines a line between two keypoints from P1 to P2.  The line
        shape may be generated as "straight" (in the active coordinate
        system) or curved.  The line shape is invariant with
        coordinate system after it is generated.  Note that solid
        modeling in a toroidal coordinate system is not recommended.
        A curved line is limited to 180 degrees.  Lines may be redefined only
        if not yet attached to an area.

        Parameters
        ----------
        p1
            Keypoint at the beginning of line.

        p2
            Keypoint at the end of line.

        ndiv
            Number of element divisions within this line.  Normally
            this field is not used; specifying divisions with LESIZE,
            etc. is recommended.

        space
            Spacing ratio.  Normally this field is not used, as
            specifying spacing ratios with the LESIZE command is
            recommended.  If positive, space is the nominal ratio of
            the last division size (at P2) to the first division size
            (at P1).  If the ratio is greater than 1, the division
            sizes increase from P1 to P2, and if less than 1, they
            decrease.  If space is negative, then ``space`` is the
            nominal ratio of the center division size to those at the
            ends.

        Returns
        -------
        int
            The line number of the generated line.

        Examples
        --------
        Create a line between the two keypoints (0, 0, 0) and (1, 0, 0)

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> lnum = mapdl.l(k0, k1)
        >>> lnum
        1

        """
        command = f"L,{p1},{p2},{ndiv},{space},{xv1},{yv1},{zv1},{xv2},{yv2},{zv2}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def l2ang(
        self, nl1="", nl2="", ang1="", ang2="", phit1="", phit2="", **kwargs
    ) -> int:
        """Generates a line at an angle with two existing lines.

        APDL Command: L2ANG

        Generates a straight line (PHIT1-PHIT2) at an angle (ANG1)
        with an existing line NL1 (P1-P2) and which is also at an
        angle (ANG2) with another existing line NL2 (P3-P4).  If the
        angles are zero the generated line is tangent to the two
        lines.  The PHIT1 and PHIT2 locations on the lines are
        automatically calculated.  Line P1-P2 becomes P1-PHIT1, P3-P4
        becomes P3-PHIT2, and new lines PHIT1-P2, PHIT2-P4, and
        PHIT1-PHIT2 are generated.  Line divisions are set to zero
        (use LESIZE, etc. to modify).

        Parameters
        ----------
        nl1
            Number of the first line to be hit (touched by the end of
            the new line).  If negative, assume P1 (see below) is the
            second keypoint of the line instead of the first.

        nl2
            Number of the second line to be hit.  If negative, assume
            P3 is the second keypoint of the line instead of the
            first.

        ang1
            Angle of intersection (usually zero or 180) of generated
            line with tangent to first line.

        ang2
            Angle of intersection (usually zero or 180) of generated
            line with tangent to second line.

        phit1
            Number to be assigned to keypoint generated at hit
            location on first line (defaults to lowest available
            keypoint number [NUMSTR]).

        phit2
            Number to be assigned to keypoint generated at hit
            location on second line (defaults to lowest available
            keypoint number [NUMSTR]).

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Create two circles and join them with a line.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 0, 1)
        >>> k2 = mapdl.k("", 0, 0, 0.5)
        >>> carc0 = mapdl.circle(k0, 1, k1)
        >>> carc1 = mapdl.circle(k2, 1, k1)
        >>> lnum = mapdl.l2ang(carc0[0], carc1[0], 90, 90)
        >>> lnum
        9

        """
        command = f"L2ANG,{nl1},{nl2},{ang1},{ang2},{phit1},{phit2}"
        msg = self.run(command, **kwargs)
        if msg:
            return parse.parse_line_no(msg)

    def l2tan(self, nl1="", nl2="", **kwargs) -> int:
        """Generates a line tangent to two lines.

        APDL Command: L2TAN

        Generates a line (P2-P3) tangent at point P2 to line NL1
        (P1-P2) and tangent at point P3 to line NL2 (P3-P4).

        Parameters
        ----------
        nl1
            Number of the first line generated line is tangent to.  If
            negative, assume P1 (see below) is the second keypoint of
            the line instead of the first.

        nl2
            Number of the second line generated line is tangent to.
            If negative, assume P3 is the second keypoint of the line
            instead of the first.

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Create two circular arcs and connect them with a spline.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 0, 1)
        >>> k2 = mapdl.k("", -1.5, 1.5, 0)
        >>> k3 = mapdl.k("", -1.5, 1.5, 1)
        >>> carc0 = mapdl.circle(k0, 1, k1, arc=90)
        >>> carc1 = mapdl.circle(k2, 1, k3, arc=90)
        >>> lnum = mapdl.l2tan(1, 2)
        3

        Plot these lines

        >>> mapdl.lplot(cpos='xy')

        """
        command = f"L2TAN,{nl1},{nl2}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def lang(self, nl1="", p3="", ang="", phit="", locat="", **kwargs) -> int:
        """Generate a straight line at an angle with a line.

        APDL Command: LANG

        Generates a straight line (PHIT-P3) at an angle (ANG) with a
        line NL1 (P1-P2).  The location of PHIT on the line is
        automatically calculated.  Line P1-P2 becomes P1-PHIT and new
        lines PHIT-P2 and PHIT-P3 are generated.  Line divisions are
        set to zero (use LESIZE, etc. to modify).

        PHIT is positioned closest to LOCAT for the given angle, ANG.
        To ensure better performance, it is recommended that LOCAT be
        input, even if it is 0.

        The program uses an iterative procedure to position PHIT.  The
        procedure is not exact, with the result that the actual value
        of ANG will sometimes differ slightly from the specified
        value.

        Parameters
        ----------
        nl1
            Number of the line to be hit (touched by the end of the
            new line).  If negative, assume P1 (see below) is the
            second keypoint of the line instead of the first.

        p3
            Keypoint at which generated line must end.

        ang
            Angle of intersection of generated line PHIT-P3 with
            tangent to line P1-P2 at PHIT.  If 0 (default), the
            generated line is tangent to NL1 toward end P1; if 90, the
            generated line is perpendicular to NL1.  If 180, the
            generated line is tangent to NL1 toward end P2.  ANG can
            be any value, but is adjusted to the corresponding acute
            angle with respect to LOCAT. See "Notes" for a discussion
            of accuracy.

        phit
            Number to be assigned to keypoint generated at hit
            location (defaults to lowest available keypoint number
            [NUMSTR]).

        locat
            Approximate location of PHIT in terms of the ratio of the
            distance along the line (NL1) to the length of the line.
            LOCAT can range from 0 to 1.  If LOCAT is blank, the point
            will be located with less speed and accuracy, and an
            arbitrary location may result.

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Create a line from a line from (0, 0, 0) to (1, 0, 0) to a
        keypoint at (1, 1, 1) at an angle of 60 degrees.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> lnum = mapdl.l(k0, k1)
        >>> k2 = mapdl.k("", 1, 1, 0)
        >>> lnum = mapdl.lang(lnum, k2, 60)
        >>> lnum
        2

        """
        command = f"LANG,{nl1},{p3},{ang},{phit},{locat}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def larc(self, p1="", p2="", pc="", rad="", **kwargs) -> int:
        """Define a circular arc.

        APDL Command: LARC

        Defines a circular arc line from P1 to P2.  The line shape is
        generated as circular, regardless of the active coordinate
        system.  The line shape is invariant with coordinate system
        after it is generated.

        When dealing with a large radius arc (1e3), or if the location
        of the arc you create is far away from the origin of your
        coordinate system, anomalies may occur. You can prevent this
        by creating the arc at a smaller scale, and then scaling the
        model back to full size (LSSCALE).

        Parameters
        ----------
        p1
            Keypoint at one end of circular arc line.

        p2
            Keypoint at other end of circular arc line.

        pc
            Keypoint defining plane of arc and center of curvature
            side (with positive radius).  Must not lie along the
            straight line from P1 to P2.  PC need not be at the center
            of curvature.

        rad
            Radius of curvature of the arc.  If negative, assume
            center of curvature side is opposite to that defined by
            PC.  If RAD is blank, RAD will be calculated from a curve
            fit through P1, PC, and P2.

        Returns
        -------
        int
            Line number of the arc.

        Examples
        --------
        Create a circular arc that travels between (0, 0, 0) and
        (1, 1, 0) with a radius of curvature of 2.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 1, 0)
        >>> k2 = mapdl.k("", 0, 1, 0)
        >>> lnum = mapdl.larc(k0, k1, k2, 2)
        1

        """
        command = f"LARC,{p1},{p2},{pc},{rad}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def larea(self, p1="", p2="", narea="", **kwargs) -> int:
        """Generate the shortest line between two keypoints on an area.

        APDL Command: LAREA

        Generates the shortest line between two keypoints, P1 and P2,
        both of which lie on an area.  The generated line will also
        lie on the area.  P1 and P2 may also be equidistant (in global
        Cartesian space) from the area (and on the same side of the
        area), in which case a line parallel to the area is generated.

        Parameters
        ----------
        p1
            First keypoint of line to be generated.

        p2
            Second keypoint of line to be generated.

        narea
            Area containing P1 and P2, or area to which generated line
            is to be parallel.

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Generate a line on a square between its two corners.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> k2 = mapdl.k("", 1, 1, 0)
        >>> k3 = mapdl.k("", 0, 1, 0)
        >>> a0 = mapdl.a(k0, k1, k2, k3)
        >>> lnum = mapdl.larea(k0, k2, a0)
        >>> lnum
        1

        """
        command = f"LAREA,{p1},{p2},{narea}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def lcomb(self, nl1="", nl2="", keep="", **kwargs) -> int:
        """Combines adjacent lines into one line.

        APDL Command: LCOMB

        Combines adjacent lines into one line (the output line).  This
        operation will effectively "undo" the LDIV operation.  Line
        divisions are set to zero (use LESIZE, etc. to modify).  Lines
        attached to the same area(s) can also be combined.  See also
        the LCCAT command for line concatenation capability.

        Parameters
        ----------
        nl1
            Number of the first line to be combined.  If NL1 = ALL,
            NL2 is ignored and all selected lines [LSEL] are combined.
            A component name may also be substituted for NL1 (NL2 is
            ignored).

        nl2
            Number of the second line to be combined.

        keep
            Specifies whether to keep the input entities:

            0 - Delete lines NL1 and NL2 and their common keypoint.
                Keypoints will not be deleted if they are meshed or if
                they are attached to other lines.  Lines will not be
                deleted if they are attached to different areas.

            1 - Keep NL1, NL2, and their common keypoint.  (The common
                keypoint will not be attached to the output line.)

        Returns
        -------
        int
            Line number of the combined line.

        Examples
        --------
        Create two lines and combine them.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> k2 = mapdl.k("", 2, 0, 0)
        >>> l0 = mapdl.l(k0, k1)
        >>> l1 = mapdl.l(k0, k2)
        >>> lout = mapdl.lcomb(l0, l1)
        >>> lout
        1

        """
        command = f"LCOMB,{nl1},{nl2},{keep}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def ldele(self, nl1="", nl2="", ninc="", kswp="", **kwargs):
        """Deletes unmeshed lines.

        APDL Command: LDELE

        Parameters
        ----------
        nl1, nl2, ninc
            Delete lines from NL1 to NL2 (defaults to NL1) in steps of NINC
            (defaults to 1).  If NL1 = ALL, NL2 and NINC are ignored and all
            selected lines [LSEL] are deleted.  If NL1 = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).  A component name may also be substituted for NL1 (NL2
            and NINC are ignored).

        kswp
            Specifies whether keypoints are also to be deleted:

            0 - Delete lines only.

            1 - Delete lines, as well as keypoints attached to lines but not attached to other
                lines.

        Notes
        -----
        A line attached to an area cannot be deleted unless the area is first
        deleted.
        """
        command = f"LDELE,{nl1},{nl2},{ninc},{kswp}"
        return self.run(command, **kwargs)

    def ldiv(self, nl1="", ratio="", pdiv="", ndiv="", keep="", **kwargs) -> str:
        """Divides a single line into two or more lines.

        APDL Command: LDIV

        Divides a single line NL1 (defined from keypoint P1 to
        keypoint P2) into two or more lines.  Line NL1 becomes the new
        line beginning with keypoint P1 and new lines are generated
        ending at keypoint P2.  If the line is attached to an area,
        the area will also be updated.  Line divisions are set to zero
        (use LESIZE, etc. to modify).

        Parameters
        ----------
        nl1
            Number of the line to be divided.  If negative, assume P1
            (see below) is the second keypoint of the line instead of
            the first for RATIO.  If ALL, divide all selected lines
            [LSEL].  A component name may also be substituted for NL1.

        ratio
            Ratio of line length P1-PDIV to line length P1-P2.  Must
            be between 0.0 and 1.0. Input ignored if NDIV > 2.

        pdiv
            Number to be assigned to keypoint generated at division
            location (defaults to lowest available keypoint number
            [NUMSTR]).  Input ignored if NL1 = ALL or NDIV > 2.  If
            PDIV already exists and lies on line NL1, divide line at
            PDIV (RATIO must also be 0.0).  If PDIV already exists and
            does not lie on line NL1, PDIV is projected and moved to
            the nearest point on line NL1 (if possible). PDIV cannot
            be attached to another line, area, or volume.

        ndiv
            The number of new lines to be generated from old line
            (defaults to 2).

        keep
            Specifies whether to keep the input entities:

            0 - Modify old line to use new keypoints and slopes.

            1 - Do not modify old line.  New lines will overlay old
                line and have unique keypoints.

        Returns
        -------
        str
            MAPDL command output.

        Examples
        --------
        Create a single line and divide it exactly half.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> l0 = mapdl.l(k0, k1)
        >>> output = mapdl.ldiv(l0, ratio=0.5)
        >>> print(output)
        DIVIDE LINE      1       RATIO=  0.50000       NEW KEYPOINT=     0
          NUMBER OF LINES DIVIDED =      1


        Create a single line and divide it into 5 pieces.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> l0 = mapdl.l(k0, k1)
        >>> mapdl.ldiv(l0, ndiv=5)

        """
        command = f"LDIV,{nl1},{ratio},{pdiv},{ndiv},{keep}"
        return self.run(command, **kwargs)

    def ldrag(
        self,
        nk1="",
        nk2="",
        nk3="",
        nk4="",
        nk5="",
        nk6="",
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        **kwargs,
    ):
        """Generates lines by sweeping a keypoint pattern along  path.

        APDL Command: LDRAG

        Parameters
        ----------
        nk1, nk2, nk3, . . . , nk6
            List of keypoints in the pattern to be dragged (6 maximum if using
            keyboard entry).  If NK1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  If
            NK1 = ALL, all selected keypoints (except those that define the
            drag path) will be swept along the path.  A component name may also
            be substituted for NK1.

        nl1, nl2, nl3, . . . , nl6
            List of lines defining the path along which the pattern is to be
            dragged (6 maximum if using keyboard entry).  Must be a continuous
            set of lines.

        Notes
        -----
        Generates lines (and their corresponding keypoints) by sweeping a given
        keypoint pattern along a characteristic drag path.  If the drag path
        consists of multiple lines, the drag direction is determined by the
        sequence in which the path lines are input (NL1, NL2, etc.).  If the
        drag path is a single line (NL1), the drag direction is from the
        keypoint on the drag line that is closest to the first keypoint of the
        given pattern to the other end of the drag line.

        The magnitude of the vector between the keypoints of the given pattern
        and the first path keypoint remains constant for all generated keypoint
        patterns and the path keypoints.  The direction of the vector relative
        to the path slope also remains constant so that patterns may be swept
        around curves.  Keypoint and line numbers are automatically assigned
        (beginning with the lowest available values [NUMSTR]).  For best
        results, the entities to be dragged should be orthogonal to the start
        of the drag path.  Drag operations that produce an error message may
        create some of the desired entities prior to terminating.
        """
        command = f"LDRAG,{nk1},{nk2},{nk3},{nk4},{nk5},{nk6},{nl1},{nl2},{nl3},{nl4},{nl5},{nl6}"
        return self.run(command, **kwargs)

    def lextnd(self, nl1="", nk1="", dist="", keep="", **kwargs) -> int:
        """Extends a line at one end by using its slope.

        APDL Command: LEXTND

        Extends a line at one end by using its slope.  Lines may be
        redefined only if not yet attached to an area.  Line divisions
        are set to zero (use LESIZE, etc. to modify).  Note that solid
        modeling in a toroidal coordinate system is not recommended.

        Parameters
        ----------
        nl1
            Number of the line to be extended.

        nk1
            Number of keypoint at the end of line NL1 to be extended.

        dist
            Distance that the line will be extended.

        keep
            Specifies whether to keep the input entities:

            0 - Modify old line to use new keypoints and slopes.

            1 - Do not modify old line.  New line will overlay old
                line and have unique keypoints.

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Create a circular arc and extend it at one of its keypoints

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 0, 1)
        >>> carcs = mapdl.circle(k0, 1, k1, arc=90)
        >>> lnum = mapdl.lextnd(carcs[0], 3, 1)
        >>> lnum
        1

        """
        command = f"LEXTND,{nl1},{nk1},{dist},{keep}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def lfillt(self, nl1="", nl2="", rad="", pcent="", **kwargs) -> int:
        """Generate a fillet line between two intersecting lines.

        APDL Command: LFILLT

        Generates a fillet line between two intersecting lines NL1
        (P1-PINT) and NL2 (P2-PINT).  Three keypoints may be
        generated, two at the fillet tangent points (PTAN1 and PTAN2)
        and one (optional) at the fillet arc center (PCENT).  Line
        P1-PINT becomes P1-PTAN1, P2-PINT becomes P2-PTAN2, and new
        arc line PTAN1-PTAN2 is generated.  Generated keypoint and
        line numbers are automatically assigned (beginning with the
        lowest available values [NUMSTR]).  Line divisions are set to
        zero (use LESIZE, etc. to modify).

        Parameters
        ----------
        nl1
            Number of the first intersecting line.

        nl2
            Number of the second intersecting line.

        rad
            Radius of fillet to be generated.  Radius should be less than the
            lengths of the two lines specified with NL1 and NL2.

        pcent
            Number to be assigned to generated keypoint at fillet arc center.
            If zero (or blank), no keypoint is generated.

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Create two intersecting lines at a right angle and add a
        fillet between them.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 1, 0)
        >>> k2 = mapdl.k("", 1, 0, 0)
        >>> l0 = mapdl.l(k0, k1)
        >>> l1 = mapdl.l(k0, k2)
        >>> lnum = mapdl.lfillt(l0, l1, 0.25)
        3

        """
        command = f"LFILLT,{nl1},{nl2},{rad},{pcent}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def lgen(
        self,
        itime="",
        nl1="",
        nl2="",
        ninc="",
        dx="",
        dy="",
        dz="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Generates additional lines from a pattern of lines.

        APDL Command: LGEN

        Parameters
        ----------
        itime
            Do this generation operation a total of ITIMEs, incrementing all
            keypoints in the given pattern automatically (or by KINC) each time
            after the first.  ITIME must be > 1 for generation to occur.

        nl1, nl2, ninc
            Generate lines from pattern beginning with NL1 to NL2 (defaults to
            NL1) in steps of NINC (defaults to 1).  If NL1 = ALL, NL2 and NINC
            are ignored and pattern is all selected lines [LSEL].  If NL1 = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NL1 (NL2 and NINC are ignored).

        dx, dy, dz
            Keypoint location increments in the active coordinate system (--,
            Dθ, DZ for cylindrical, --, Dθ, -- for spherical).

        kinc
            Keypoint increment between generated sets.  If zero, the lowest
            available keypoint numbers are assigned [NUMSTR].

        noelem
            Specifies if elements and nodes are also to be generated:

            0 - Generate nodes and elements associated with the original lines, if they exist.

            1 - Do not generate nodes and elements.

        imove
            Specifies whether to redefine the existing lines:

            0 - Generate additional lines as requested with the ITIME argument.

            1 - Move original lines to new position retaining the same keypoint numbers (ITIME,
                KINC, and NOELM are ignored).  Valid only if the old lines are
                no longer needed at their original positions.  Corresponding
                meshed items are also moved if not needed at their original
                position.

        Notes
        -----
        Generates additional lines (and their corresponding keypoints and mesh)
        from a given line pattern.  The MAT, TYPE, REAL, and ESYS attributes
        are based upon the lines in the pattern and not upon the current
        settings.  End slopes of the generated lines remain the same (in the
        active coordinate system) as those of the given pattern.  For example,
        radial slopes remain radial, etc.  Generations which produce lines of a
        size or shape different from the pattern (i.e., radial generations in
        cylindrical systems, radial and phi generations in spherical systems,
        and theta generations in elliptical systems) are not allowed.  Note
        that solid modeling in a toroidal coordinate system is not recommended.
        New line numbers are automatically assigned (beginning with the lowest
        available values [NUMSTR]).
        """
        command = (
            f"LGEN,{itime},{nl1},{nl2},{ninc},{dx},{dy},{dz},{kinc},{noelem},{imove}"
        )
        return self.run(command, **kwargs)

    def llist(self, nl1="", nl2="", ninc="", lab="", **kwargs):
        """Lists the defined lines.

        APDL Command: LLIST

        Parameters
        ----------
        nl1, nl2, ninc
            List lines from NL1 to NL2 (defaults to NL1) in steps of NINC
            (defaults to 1).  If NL1 = ALL (default), NL2 and NINC are ignored
            and all selected lines [LSEL] are listed.  If NL1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NL1 (NL2 and NINC are ignored).

        lab
            Determines what type of listing is used (one of the following):

            (blank) - Prints information about all lines in the specified range.

            RADIUS - Prints the radius of certain circular arcs, along with the keypoint numbers of
                     each line.  Straight lines, non-circular curves, and
                     circular arcs not internally identified as arcs (which
                     depends upon how each arc is created) will print a radius
                     value of zero.

            LAYER - Prints layer-mesh control specifications.

            HPT - Prints information about only those lines that contain hard points. HPT is not
                  supported in the GUI.

            ORIENT - Prints a list of lines, and identifies any orientation keypoints and any cross
                     section IDs that are associated with the lines.  Used for
                     beam meshing with defined orientation nodes and cross
                     sections.

        Notes
        -----
        There are 2 listings for the number of element divisions and the
        spacing ratio. The first listing shows assignments from LESIZE only,
        followed by the "hard" key (KYNDIV). See LESIZE for more information.
        The second listing shows NDIV and SPACE for the existing mesh, if one
        exists. Whether this existing mesh and the mesh generated by LESIZE
        match at any given point depends upon meshing options and the sequence
        of meshing operations.

        A "-1" in the "nodes" column indicates that the line has been meshed
        but that there are no interior nodes.

        An attribute (TYPE, MAT, REAL, or ESYS) listed as a zero is unassigned;
        one listed as a positive value indicates that the attribute was
        assigned with the LATT command (and will not be reset to zero if the
        mesh is cleared); one listed as a negative value indicates that the
        attribute was assigned using the attribute pointer [TYPE, MAT, REAL, or
        ESYS] that was active during meshing (and will be reset to zero if the
        mesh is cleared).

        This command is valid in any processor.
        """
        command = f"LLIST,{nl1},{nl2},{ninc},{lab}"
        return self.run(command, **kwargs)

    def lplot(self, nl1="", nl2="", ninc="", **kwargs):
        """Displays the selected lines.

        APDL Command: LPLOT

        Parameters
        ----------
        nl1, nl2, ninc
            Display lines from NL1 to NL2 (defaults to NL1) in steps of NINC
            (defaults to 1).  If NL1 = ALL (default), NL2 and NINC are ignored
            and display all selected lines [LSEL].

        Notes
        -----
        Mesh divisions on plotted lines are controlled by the LDIV option of
        the /PSYMB command.

        This command is valid in any processor.
        """
        command = f"LPLOT,{nl1},{nl2},{ninc}"
        return self.run(command, **kwargs)

    def lreverse(self, lnum="", noeflip="", **kwargs):
        """Reverses the normal of a line, regardless of its connectivity or mesh

        APDL Command: LREVERSE
        status.

        Parameters
        ----------
        lnum
            Line number of the line whose normal direction is to be reversed.
            If LNUM = ALL, the normals of all selected lines will be reversed.
            If LNUM = P, graphical picking is enabled.  A component name may
            also be substituted for LNUM.

        noeflip
            Indicates whether you want to change the normal direction of the
            existing elements on the reversed line(s) so that they are
            consistent with each line's new normal direction.

            0 - Make the normal direction of existing elements on the reversed line(s)
                consistent with each line's new normal direction (default).

            1 - Do not change the normal direction of existing elements on the reversed
                line(s).

        Notes
        -----
        You cannot use the LREVERSE command to change the normal direction of
        any element that has a body or surface load.  We recommend that you
        apply all of your loads only after ensuring that the element normal
        directions are acceptable.

        Real constants (such as nonuniform shell thickness and tapered beam
        constants) may be invalidated by an element reversal.

        For more information, see Revising Your Model in the Modeling and
        Meshing Guide.
        """
        command = f"LREVERSE,{lnum},{noeflip}"
        return self.run(command, **kwargs)

    def lrotat(
        self,
        nk1="",
        nk2="",
        nk3="",
        nk4="",
        nk5="",
        nk6="",
        pax1="",
        pax2="",
        arc="",
        nseg="",
        **kwargs,
    ):
        """Generates circular lines by rotating a keypoint pattern about an axis.

        APDL Command: LROTAT

        Parameters
        ----------
        nk1, nk2, nk3, . . . , nk6
            List of keypoints in the pattern to be rotated (6 maximum if using
            keyboard entry).  If NK1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  If
            NK1 = ALL, all selected keypoints (except PAX1 and PAX2) will
            define the pattern to be rotated.  A component name may also be
            substituted for NK1.

        pax1, pax2
            Keypoints defining the axis about which the keypoint pattern is to
            be rotated.

        arc
            Arc length (in degrees).  Positive follows right-hand rule about
            PAX1-PAX2 vector.  Defaults to 360.

        nseg
            Number of lines (8 maximum) around circumference.  Defaults to
            minimum required for 90° (maximum) arcs, i.e., 4 for 360°, 3 for
            270°, etc.

        Notes
        -----
        Generates circular lines (and their corresponding keypoints) by
        rotating a keypoint pattern about an axis.  Keypoint patterns are
        generated at regular angular locations (based on a maximum spacing of
        90°).  Line patterns are generated at the keypoint patterns.  Keypoint
        and line numbers are automatically assigned (beginning with the lowest
        available values [NUMSTR]).
        """
        command = (
            f"LROTAT,{nk1},{nk2},{nk3},{nk4},{nk5},{nk6},{pax1},{pax2},{arc},{nseg}"
        )
        return self.run(command, **kwargs)

    def lsscale(
        self,
        nl1="",
        nl2="",
        ninc="",
        rx="",
        ry="",
        rz="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Generates a scaled set of lines from a pattern of lines.

        APDL Command: LSSCALE

        Parameters
        ----------
        nl1, nl2, ninc
            Set of lines (NL1 to NL2 in steps of NINC) that defines the pattern
            to be scaled.  NL2 defaults to NL1, NINC defaults to 1.  If NL1 =
            ALL, NL2 and NINC are ignored and the pattern is defined by all
            selected lines.  If NL1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  A
            component name may also be substituted for NL1 (NL2 and NINC are
            ignored).

        rx, ry, rz
            Scale factors to be applied to the X, Y, Z keypoint coordinates in
            active coordinate system (RR, Rθ, RZ for cylindrical; RR, Rθ, RΦ
            for spherical).  Note that the Rθ and RΦ scale factors are
            interpreted as angular offsets.  For example, for CSYS,1, RR, Rθ,
            RZ input of (1.5,10,3) would scale the specified keypoints 1.5
            times in the radial and 3 times in the Z direction, while adding an
            offset of 10 degrees to the keypoints.  Zero, blank, or negative
            scale factor values are assumed to be 1.0.  Zero or blank angular
            offsets have no effect.

        kinc
            Increment to be applied to keypoint numbers for generated set.  If
            zero, the lowest available keypoint numbers will be assigned
            [NUMSTR].

        noelem
            Specifies whether nodes and elements are also to be generated:

            0 - Nodes and line elements associated with the original lines will be generated
                (scaled) if they exist.

            1 - Nodes and line elements will not be generated.

        imove
            Specifies whether lines will be moved or newly defined:

            0 - Additional lines will be generated.

            1 - Original lines will be moved to new position (KINC and NOELEM are ignored).
                Use only if the old lines are no longer needed at their
                original positions.  Corresponding meshed items are also moved
                if not needed at their original position.

        Notes
        -----
        Generates a scaled set of lines (and their corresponding keypoints and
        mesh) from a pattern of lines.  The MAT, TYPE, REAL, and ESYS
        attributes are based on the lines in the pattern and not the current
        settings.  Scaling is done in the active coordinate system.  Lines in
        the pattern could have been generated in any coordinate system.
        """
        command = f"LSSCALE,{nl1},{nl2},{ninc},{rx},{ry},{rz},{kinc},{noelem},{imove}"
        return self.run(command, **kwargs)

    def lstr(self, p1="", p2="", **kwargs) -> int:
        """Define a straight line irrespective of the active coordinate system.

        APDL Command: LSTR

        Defines a straight line from P1 to P2 using the global
        Cartesian coordinate system.  The active coordinate system
        will be ignored.  The line shape is invariant with the
        coordinate system after it is generated.  Lines may be
        redefined only if not yet attached to an area.

        Parameters
        ----------
        p1
            Keypoint at the beginning of line.

        p2
            Keypoint at the end of line.

        Returns
        -------
        int
            Line number of the generated line.

        Examples
        --------
        Create a cartesian straight line regardless of the coordinate
        system being in cylindrical.

        >>> mapdl.csys(1)
        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 1, 1)
        >>> lnum = mapdl.lstr(k0, k1)
        >>> lnum
        1

        """
        command = f"LSTR,{p1},{p2}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def lsum(self, **kwargs):
        """Calculates and prints geometry statistics of the selected lines.

        APDL Command: LSUM

        Notes
        -----
        Calculates and prints geometry statistics (length, centroid, moments of
        inertia, etc.) associated with the selected lines.  Geometry items are
        reported in the global Cartesian coordinate system.  A unit density is
        assumed, irrespective of any material associations [LATT, MAT].  Items
        calculated by LSUM and later retrieved by a ``*GET`` or ``*VGET`` command are
        valid only if the model is not modified after the LSUM command is
        issued.
        """
        command = f"LSUM,"
        return self.run(command, **kwargs)

    def lsymm(
        self,
        ncomp="",
        nl1="",
        nl2="",
        ninc="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Generates lines from a line pattern by symmetry reflection.

        APDL Command: LSYMM

        Parameters
        ----------
        ncomp
            Symmetry key:

            X - X symmetry (default).

            Y - Y symmetry.

            Z - Z symmetry.

        nl1, nl2, ninc
            Reflect lines from pattern beginning with NL1 to NL2 (defaults to
            NL1) in steps of NINC (defaults to 1).  If NL1 = ALL, NL2 and NINC
            are ignored and pattern is all selected lines [LSEL].  If NL1 = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NL1 (NL2 and NINC are ignored).

        kinc
            Keypoint increment between sets.  If zero, the lowest available
            keypoint  numbers are assigned [NUMSTR].

        noelem
            Specifies whether nodes and elements are also to be generated:

            0 - Generate nodes and elements associated with the original lines, if they exist.

            1 - Do not generate nodes and elements.

        imove
            Specifies whether areas will be moved or newly defined:

            0 - Generate additional lines.

            1 - Move original lines to new position retaining the same keypoint numbers (KINC
                and NOELEM are ignored).  Valid only if the old lines are no
                longer needed at their original positions.  Corresponding
                meshed items are also moved if not needed at their original
                position.

        Notes
        -----
        Generates a reflected set of lines (and their corresponding keypoints
        and mesh) from a given line pattern by a symmetry reflection (see
        analogous node symmetry command, NSYM).  The MAT, TYPE, REAL, and ESYS
        attributes are based upon the lines in the pattern and not upon the
        current settings.  Reflection is done in the active coordinate system
        by changing a particular coordinate sign.  The active coordinate system
        must be Cartesian.  Lines in the pattern may have been generated in any
        coordinate system.  However, solid modeling in a toroidal coordinate
        system is not recommended.  Lines are generated as described in the
        LGEN command.

        See the ESYM command for additional information about symmetry
        elements.
        """
        command = f"LSYMM,{ncomp},{nl1},{nl2},{ninc},{kinc},{noelem},{imove}"
        return self.run(command, **kwargs)

    def ltan(self, nl1="", p3="", xv3="", yv3="", zv3="", **kwargs) -> int:
        """Generate a line at the end of, and tangent to, an existing line.

        APDL Command: LTAN

        Generates a line (P2-P3) tangent at end point (P2) of line NL1
        (P1-P2).

        Parameters
        ----------
        nl1
            Number of the line the generated line is tangent to.  If
            negative, assume P1 (see below), instead of P2, is the
            second keypoint of line NL1.

        p3
            Keypoint at which generated line must end.

        Returns
        -------
        int
            Line number of the line generated.

        Examples
        --------
        Create a circular arc and generate a tangent spline at the end
        of it directed to a new keypoint.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 0, 1)
        >>> k2 = mapdl.k("", -1, 1.5, 0)
        >>> carc = mapdl.circle(k0, 1, k1, arc=90)
        >>> lnum = mapdl.ltan(carc[0], k2)
        >>> lnum
        2

        """
        command = f"LTAN,{nl1},{p3},{xv3},{yv3},{zv3}"
        return parse.parse_line_no(self.run(command, **kwargs))

    def ltran(
        self,
        kcnto="",
        nl1="",
        nl2="",
        ninc="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Transfers a pattern of lines to another coordinate system.

        APDL Command: LTRAN

        Parameters
        ----------
        kcnto
            Reference number of coordinate system where the pattern is to be
            transferred.  Transfer occurs from the active coordinate system.
            The coordinate system type and parameters of KCNTO must be the same
            as the active system.

        nl1, nl2, ninc
            Transfer lines from pattern beginning with NL1 to NL2 (defaults to
            NL1) in steps of NINC (defaults to 1).  If NL1 = ALL, NL2 and NINC
            are ignored and pattern is all selected lines [LSEL].  If NL1 = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NL1 (NL2 and NINC are ignored).

        kinc
            Keypoint increment between sets.  If zero, the lowest available
            keypoint numbers are assigned [NUMSTR].

        noelem
            Specifies whether nodes and elements are also to be generated:

            0 - Generate nodes and elements associated with the original lines, if they exist.

            1 - Do not generate nodes and elements.

        imove
            Specifies whether lines will be moved or newly defined:

            0 - Generate additional lines.

            1 - Move original lines to new position retaining the same keypoint numbers (KINC
                and NOELM are ignored).  Valid only if the old lines are no
                longer needed at their original positions.  Corresponding
                meshed items are also moved if not needed at their original
                position.

        Notes
        -----
        Transfers a pattern of lines (and their corresponding keypoints and
        mesh) from one coordinate system to another (see analogous node
        transfer command, TRANSFER).  The MAT, TYPE, REAL, and ESYS attributes
        are based upon the lines in the pattern and not upon the current
        settings.  Coordinate systems may be translated and rotated relative to
        each other.  Initial pattern may be generated in any coordinate system.
        However, solid modeling in a toroidal coordinate system is not
        recommended.  Coordinate and slope values are interpreted in the active
        coordinate system and are transferred directly.  Lines are generated as
        described in the LGEN command.
        """
        command = f"LTRAN,{kcnto},{nl1},{nl2},{ninc},{kinc},{noelem},{imove}"
        return self.run(command, **kwargs)

    def spline(
        self,
        p1="",
        p2="",
        p3="",
        p4="",
        p5="",
        p6="",
        xv1="",
        yv1="",
        zv1="",
        xv6="",
        yv6="",
        zv6="",
        **kwargs,
    ) -> list:
        """Generate a segmented spline through a series of keypoints.

        APDL Command: SPLINE

        The output from this command is a series of connected lines
        (one line between each pair of keypoints) that together form a
        spline.  Note that solid modeling in a toroidal coordinate
        system is not recommended.

        Parameters
        ----------
        p1, p2, p3, . . . , p6
            Keypoints through which the spline is fit.  At least two
            must be defined.

        Returns
        -------
        list
            List of line numbers generated.

        Examples
        --------
        Create a spline with 5 keypoints.

        >>> k0 = mapdl.k('', 0, 0, 0)
        >>> k1 = mapdl.k('', 0.2, 0.2, 0)
        >>> k2 = mapdl.k('', 0.4, 0.3, 0)
        >>> k3 = mapdl.k('', 0.6, 0.5, 0)
        >>> k4 = mapdl.k('', 0.8, 0.3, 0)
        >>> lines = mapdl.spline(k0, k1, k2, k3, k4)
        >>> lines
        [1, 2, 3, 4]

        """
        command = (
            f"SPLINE,{p1},{p2},{p3},{p4},{p5},{p6},{xv1},{yv1},{zv1},{xv6},{yv6},{zv6}"
        )
        return parse.parse_line_nos(self.run(command, **kwargs))

    def ssln(self, fact="", size="", **kwargs):
        """Selects and displays small lines in the model.

        APDL Command: SSLN

        Parameters
        ----------
        fact
            Factor used to determine small lines.  FACT times the average line
            length in the model is used as the line length limit below which
            lines will be selected.

        size
            Line length limit for line selection.  Lines that have a length
            less than or equal to SIZE will be selected.  Used only if FACT is
            blank.

        Notes
        -----
        SSLN invokes a predefined ANSYS macro for selecting small lines in a
        model.  Lines that are smaller than or equal to the specified limit
        (FACT or SIZE) are selected and line numbers are displayed.  This
        command macro is useful for detecting very small lines in a model that
        may cause problems (i.e., poorly shaped elements or a meshing failure)
        during meshing.  All lines that are not "small" will be unselected and
        can be reselected with the LSEL command.
        """
        command = f"SSLN,{fact},{size}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import (
    all_others,
    analysis_options,
    birth_and_death,
    dynamic_options,
    fe_body_loads,
    fe_constraints,
    fe_forces,
    fe_surface_loads,
    gap_conditions,
    inertia,
    load_step_operations,
    load_step_options,
    master_dof,
    miscellaneous_loads,
    multi_field_solver_convergence_controls,
    multi_field_solver_definition_commands,
    multi_field_solver_global_controls,
    multi_field_solver_interface_mapping,
    multi_field_solver_load_transfer,
    multi_field_solver_time_controls,
    nonlinear_options,
    ocean,
    radiosity,
    rezoning,
    solid_body_loads,
    solid_constraints,
    solid_forces,
    solid_surface_loads,
    solution_status,
    spectrum_options,
    twod_to_3d_analysis,
)
